"""
Device Tree Blob (or Flattened Device Tree) OFRAK Utilities
For more information see: https://devicetree-specification.readthedocs.io/en/stable/flattened-format.html
"""

import os
import struct
from dataclasses import dataclass
from enum import Enum
from typing import Union, List, Tuple

import fdt

from ofrak.component.analyzer import Analyzer
from ofrak.component.identifier import Identifier
from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.model.viewable_tag_model import AttributesType
from ofrak.resource import Resource
from ofrak.service.resource_service_i import ResourceFilter, ResourceSort
from ofrak.core import GenericBinary, MagicMimeIdentifier, MagicDescriptionIdentifier
from ofrak.model.component_model import ComponentConfig
from ofrak.model.resource_model import index
from ofrak_type.range import Range

DTB_MAGIC_SIGNATURE: int = 0xD00DFEED


@dataclass
class DtbNode(GenericBinary):
    """
    Device Tree Node
    """

    name: str

    @index
    def DtbNodeName(self) -> str:
        return self.name

    @classmethod
    def caption(cls, attributes) -> str:
        try:
            dtb_attributes = attributes[AttributesType[DtbNode]]
        except KeyError:
            return super().caption(attributes)
        return f"{cls.__name__}: {dtb_attributes.name}"

    async def get_path(self) -> str:
        """
        Get the path of a DtbNode within a DeviceTreeBlob.
        Root node is always "/" per DTB specifications.
        """
        if self.name == "/":
            return self.name

        parent_node = await self.resource.get_parent_as_view(v_type=DtbNode)
        return os.path.join(await parent_node.get_path(), self.name)


class DeviceTreeBlob(GenericBinary):
    """
    A Device Tree Blob (DTB).
    """

    async def get_node_by_path(self, path: str) -> DtbNode:
        descendants = await self.resource.get_descendants_as_view(
            v_type=DtbNode, r_filter=ResourceFilter.with_tags(DtbNode)
        )
        for node in descendants:
            d_path = await node.get_path()
            if d_path == path:
                return node
        raise ValueError(f"The path {path} does not correspond to a node")


@dataclass
class DtbHeader(GenericBinary):
    """
    Device Tree Header
    """

    dtb_magic: int
    totalsize: int
    off_dt_struct: int
    off_dt_strings: int
    off_mem_rsvmap: int
    version: int
    last_comp_version: int
    boot_cpuid_phys: int = 0
    size_dt_strings: int = 0
    size_dt_struct: int = 0


class DtbHeaderAnalyzer(Analyzer[None, DtbHeader]):
    """
    Analyze Device Tree Blob header information and return a DtbHeader
    """

    id = b"DtbHeaderAnalyzer"
    targets = (DtbHeader,)
    outputs = (DtbHeader,)

    async def analyze(self, resource: Resource, config: None) -> DtbHeader:
        header_data = await resource.get_data()
        (
            dtb_magic,
            totalsize,
            off_dt_struct,
            off_dt_strings,
            off_mem_rsvmap,
            version,
            last_comp_version,
        ) = struct.unpack(">IIIIIII", header_data[:28])
        assert dtb_magic == DTB_MAGIC_SIGNATURE, (
            f"DTB Magic bytes not matching."
            f"Expected: {DTB_MAGIC_SIGNATURE} "
            f"Unpacked: {dtb_magic}"
        )
        boot_cpuid_phys = 0
        dtb_strings_size = 0
        dtb_struct_size = 0
        if version >= 2:
            boot_cpuid_phys = struct.unpack(">I", header_data[28:32])[0]
        if version >= 3:
            dtb_strings_size = struct.unpack(">I", header_data[32:36])[0]
        if version >= 17:
            dtb_struct_size = struct.unpack(">I", header_data[36:40])[0]

        return DtbHeader(
            dtb_magic,
            totalsize,
            off_dt_struct,
            off_dt_strings,
            off_mem_rsvmap,
            version,
            last_comp_version,
            boot_cpuid_phys,
            dtb_strings_size,
            dtb_struct_size,
        )


@dataclass
class DtbEntry(GenericBinary):
    """
    Device Tree Entry
    """

    address: int
    size: int


class DtbPropertyType(Enum):
    DtbPropNoValue = 0
    DtbInt = 1
    DtbStr = 2
    DtbBytes = 3
    DtbIntList = 4
    DtbStrList = 5


@dataclass
class DtbProperty(GenericBinary):
    """
    DTB Property
    """

    name: str
    p_type: DtbPropertyType

    @index
    def DtbPropertyName(self) -> str:
        return self.name

    @classmethod
    def caption(cls, attributes) -> str:
        try:
            dtb_attributes = attributes[AttributesType[DtbProperty]]
        except KeyError:
            return super().caption(attributes)
        return f"{cls.__name__}: {dtb_attributes.name}"

    async def get_path(self):
        parent_node = await self.resource.get_parent_as_view(v_type=DtbNode)
        return os.path.join(await parent_node.get_path(), self.name)

    async def get_value(self) -> Union[str, List[str], int, List[int], bytes, bytearray, None]:
        if self.p_type is DtbPropertyType.DtbPropNoValue:
            return None
        elif self.p_type is DtbPropertyType.DtbBytes:
            return await self.resource.get_data()
        elif self.p_type is DtbPropertyType.DtbInt:
            return struct.unpack(">I", await self.resource.get_data())[0]
        elif self.p_type is DtbPropertyType.DtbIntList:
            data = await self.resource.get_data()
            return [
                struct.unpack(">I", i)[0] for i in [data[j : j + 4] for j in range(0, len(data), 4)]
            ]
        elif self.p_type is DtbPropertyType.DtbStr:
            data = await self.resource.get_data()
            return data.decode("ascii")
        elif self.p_type is DtbPropertyType.DtbStrList:
            data = await self.resource.get_data()
            return [s.decode("ascii") for s in data.split(b"\x00")]
        else:
            raise TypeError(f"Unsupported type {self.p_type} for property {self.name}")


class DeviceTreeBlobUnpacker(Unpacker[None]):
    """
    Unpacks a DeviceTreeBlob:

    A DeviceTreeBlob consists of:
    - 1 DtbHeader
    - 0 or more DtbEntry instances
    - 1 root DtbNode which can contain 0 or more DtbNode and DtbProperty children
    - Each DtbNode can have 0 or more DtbProperty children and 0 or more further nested DtbNode
        children in it
    """

    targets = (DeviceTreeBlob,)
    children = (
        DtbHeader,
        DtbEntry,
        DtbNode,
        DtbProperty,
    )

    async def unpack(self, resource: Resource, config: ComponentConfig = None):
        dtb_data = await resource.get_data()
        dtb_view = await resource.view_as(DeviceTreeBlob)
        dtb = fdt.parse_dtb(dtb_data)

        # Create DtbHeader
        await resource.create_child(
            tags=(DtbHeader,),
            data=dtb.header.export(),
        )

        # Create DtbEntry instances
        for dtb_entry in dtb.entries:
            await resource.create_child_from_view(
                DtbEntry(
                    address=dtb_entry["address"],
                    size=dtb_entry["size"],
                ),
                data=b"",
            )

        # Create root node
        await resource.create_child_from_view(
            DtbNode(name=dtb.root.name),
            data=b"",
        )

        # Create DtbNode and DtbProperty instances and structure by walking the DeviceTreeBlob
        for path, nodes, props in dtb.walk():
            # Get parent
            parent_node = await dtb_view.get_node_by_path(path)
            for node in nodes:
                await parent_node.resource.create_child_from_view(DtbNode(name=node.name), data=b"")
            for prop in props:
                p_type, p_data = _prop_from_fdt(prop)

                await parent_node.resource.create_child_from_view(
                    DtbProperty(
                        name=prop.name,
                        p_type=p_type,
                    ),
                    data=p_data,
                )


class DeviceTreeBlobPacker(Packer[None]):
    """
    Device Tree Blob Packer

    Repacks the Device Tree Blob tree structure into the binary format and patches the original
    resource.
    """

    id = b"DeviceTreeBlobPacker"
    targets = (DeviceTreeBlob,)

    async def pack(self, resource: Resource, config: ComponentConfig = None):
        header = fdt.Header()
        header_view = await resource.get_only_descendant_as_view(
            v_type=DtbHeader, r_filter=ResourceFilter(tags=[DtbHeader])
        )

        header.version = header_view.version
        header.total_size = header_view.totalsize
        header.off_dt_struct = header_view.off_dt_struct
        header.last_comp_version = header_view.last_comp_version
        header.boot_cpuid_phys = header_view.boot_cpuid_phys

        dtb = fdt.FDT(header=header)

        dtb.entries = [
            {"address": entry.address, "size": entry.size}
            for entry in await resource.get_descendants_as_view(
                v_type=DtbEntry, r_filter=ResourceFilter(tags=[DtbEntry])
            )
        ]

        root_node_view = await resource.get_only_child_as_view(
            DtbNode, r_filter=ResourceFilter(tags=[DtbNode])
        )
        dtb.root = fdt.Node(name=await root_node_view.get_path())
        for prop in await root_node_view.resource.get_children_as_view(
            v_type=DtbProperty,
            r_filter=ResourceFilter(tags=[DtbProperty]),
            r_sort=ResourceSort(DtbProperty.DtbPropertyName),
        ):
            dtb.add_item(await _prop_to_fdt(prop), await root_node_view.get_path())
        for node in await root_node_view.resource.get_descendants_as_view(
            v_type=DtbNode,
            r_filter=ResourceFilter(tags=[DtbNode]),
            r_sort=ResourceSort(DtbNode.DtbNodeName),
        ):
            # By default, add_item adds the missing nodes to complete the path of a previous node
            if not dtb.exist_node(await node.get_path()):
                dtb.add_item(fdt.Node(node.name), os.path.dirname(await node.get_path()))
            for prop in await node.resource.get_children_as_view(
                v_type=DtbProperty,
                r_filter=ResourceFilter(tags=[DtbProperty]),
                r_sort=ResourceSort(DtbProperty.DtbPropertyName),
            ):
                dtb.add_item(await _prop_to_fdt(prop), await node.get_path())
        original_size = await resource.get_data_length()
        resource.queue_patch(Range(0, original_size), dtb.to_dtb())


class DeviceTreeBlobIdentifier(Identifier[None]):
    """
    Identify Device Tree Blob files.
    """

    targets = (GenericBinary,)

    async def identify(self, resource: Resource, config: ComponentConfig = None) -> None:
        """
        Identify DTB files based on the first four bytes being "d00dfeed".
        """
        data = await resource.get_data(Range(0, 4))
        if data == struct.pack("<I", DTB_MAGIC_SIGNATURE):
            resource.add_tag(DeviceTreeBlob)


async def _prop_to_fdt(p: DtbProperty) -> fdt.items.Property:
    """
    Generates an fdt.items.property corresponding to a DtbProperty.
    :param p:
    :return:
    """
    value = await p.get_value()
    if p.p_type is DtbPropertyType.DtbPropNoValue:
        return fdt.items.Property(name=p.name)
    elif p.p_type is DtbPropertyType.DtbBytes:
        return fdt.items.PropBytes(name=p.name, data=await p.resource.get_data())
    elif p.p_type is DtbPropertyType.DtbInt:
        return fdt.items.PropWords(p.name, value)
    elif p.p_type is DtbPropertyType.DtbIntList:
        return fdt.items.PropWords(p.name, *value)
    elif p.p_type is DtbPropertyType.DtbStr:
        return fdt.items.PropStrings(p.name, value)
    elif p.p_type is DtbPropertyType.DtbStrList:
        return fdt.items.PropStrings(p.name, *value)
    else:
        raise TypeError(f"Unsupported type {p.p_type} for property {p.name}")


def _prop_from_fdt(p: fdt.items.Property) -> Tuple[DtbPropertyType, bytes]:
    """
    Converts an fdt.items.property to its p_type and p_data values.
    :param p:
    :return:
    """
    if type(p) is fdt.items.Property or len(p.data) == 0:
        _p_type = DtbPropertyType.DtbPropNoValue
        _p_data = b""
    elif type(p) is fdt.items.PropBytes:
        _p_type = DtbPropertyType.DtbBytes
        _p_data = bytes(p.data)
    elif isinstance(p.value, int):
        if len(p.data) == 1:
            _p_type = DtbPropertyType.DtbInt
        else:
            _p_type = DtbPropertyType.DtbIntList
        _p_data = b"".join([struct.pack(">I", i) for i in p.data])
    elif isinstance(p.value, str):
        if len(p.data) == 1:
            _p_type = DtbPropertyType.DtbStr
            _p_data = b"".join([s.encode("ascii") for s in p.data])
        else:
            _p_type = DtbPropertyType.DtbStrList
            _p_data = b"\0".join([s.encode("ascii") for s in p.data])

    else:
        raise TypeError(f"Unknown type for DTB Property: {p}")
    return _p_type, _p_data


MagicMimeIdentifier.register(DeviceTreeBlob, "Device Tree Blob")
MagicDescriptionIdentifier.register(DeviceTreeBlob, lambda s: "device tree blob" in s.lower())



import asyncio
import logging
import tempfile
from dataclasses import dataclass
from subprocess import CalledProcessError

from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.resource import Resource
from ofrak.core.filesystem import File, Folder, FilesystemRoot, SpecialFileType

from ofrak.core.magic import MagicDescriptionIdentifier
from ofrak.core.binary import GenericBinary
from ofrak_type.range import Range
from ofrak.model.component_model import ComponentExternalTool

LOGGER = logging.getLogger(__name__)

JEFFERSON = ComponentExternalTool("jefferson", "https://pypi.org/project/jefferson/", "--help")

MKFS_JFFS2 = ComponentExternalTool(
    "mkfs.jffs2", "http://linux-mtd.infradead.org/", "-help", "mtd-utils"
)


@dataclass
class Jffs2Filesystem(GenericBinary, FilesystemRoot):
    """
    Filesystem stored in a JFFS2 format.
    """


class Jffs2Unpacker(Unpacker[None]):
    """Unpack a JFFS2 filesystem."""

    targets = (Jffs2Filesystem,)
    children = (File, Folder, SpecialFileType)
    external_dependencies = (JEFFERSON,)

    async def unpack(self, resource: Resource, config=None):
        with tempfile.NamedTemporaryFile() as temp_file:
            resource_data = await resource.get_data()
            temp_file.write(resource_data)
            temp_file.flush()

            with tempfile.TemporaryDirectory() as temp_flush_dir:
                cmd = [
                    "jefferson",
                    "--force",
                    "--dest",
                    temp_flush_dir,
                    temp_file.name,
                ]
                proc = await asyncio.create_subprocess_exec(
                    *cmd,
                )
                returncode = await proc.wait()
                if proc.returncode:
                    raise CalledProcessError(returncode=returncode, cmd=cmd)

                jffs2_view = await resource.view_as(Jffs2Filesystem)
                await jffs2_view.initialize_from_disk(temp_flush_dir)


class Jffs2Packer(Packer[None]):
    """
    Pack files into a compressed JFFS2 filesystem.
    """

    targets = (Jffs2Filesystem,)
    external_dependencies = (MKFS_JFFS2,)

    async def pack(self, resource: Resource, config=None):
        jffs2_view: Jffs2Filesystem = await resource.view_as(Jffs2Filesystem)
        temp_flush_dir = await jffs2_view.flush_to_disk()
        with tempfile.NamedTemporaryFile(suffix=".sqsh", mode="rb") as temp:
            cmd = [
                "mkfs.jffs2",
                "-r",
                temp_flush_dir,
                "-o",
                temp.name,
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
            )
            returncode = await proc.wait()
            if proc.returncode:
                raise CalledProcessError(returncode=returncode, cmd=cmd)
            new_data = temp.read()
            # Passing in the original range effectively replaces the original data with the new data
            resource.queue_patch(Range(0, await resource.get_data_length()), new_data)


MagicDescriptionIdentifier.register(Jffs2Filesystem, lambda s: "jffs2 filesystem" in s.lower())



import asyncio
import tempfile
from dataclasses import dataclass
from typing import Dict, Optional

from ofrak.component.analyzer import Analyzer
from ofrak.resource import Resource
from ofrak.model.component_model import ComponentConfig, ComponentExternalTool
from ofrak.model.resource_model import ResourceAttributes


@dataclass
class StringsAnalyzerConfig(ComponentConfig):
    min_length: int = 8


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class StringsAttributes(ResourceAttributes):
    strings: Dict[int, str]


class _StringsToolDependency(ComponentExternalTool):
    def __init__(self):
        super().__init__(
            "strings",
            "https://linux.die.net/man/1/strings",
            "--help",
            apt_package="binutils",
            brew_package="binutils",
        )

    async def is_tool_installed(self) -> bool:
        try:
            cmd = [
                self.tool,
                self.install_check_arg,
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL,
            )

            # ignore returncode because "strings --help" on Mac has returncode 1
            await proc.wait()
        except FileNotFoundError:
            return False

        return True


class StringsAnalyzer(Analyzer[Optional[StringsAnalyzerConfig], StringsAttributes]):
    targets = ()
    outputs = (StringsAttributes,)
    external_dependencies = (_StringsToolDependency(),)

    async def analyze(
        self, resource: Resource, config: Optional[StringsAnalyzerConfig] = None
    ) -> StringsAttributes:
        if config is None:
            config = StringsAnalyzerConfig()

        strings = dict()
        with tempfile.NamedTemporaryFile() as temp_file:
            temp_file.write(await resource.get_data())
            temp_file.flush()

            proc = await asyncio.subprocess.create_subprocess_exec(
                "strings",
                "-t",
                "d",
                f"-{config.min_length}",
                temp_file.name,
                stdout=asyncio.subprocess.PIPE,
            )

            line = await proc.stdout.readline()  # type: ignore
            while line:
                line = line.decode("ascii").strip()
                try:
                    offset, string = line.split(" ", maxsplit=1)
                except ValueError as e:
                    # String consisted entirely of whitespace
                    line = await proc.stdout.readline()  # type: ignore
                    continue
                strings[int(offset)] = string
                line = await proc.stdout.readline()  # type: ignore
            await proc.wait()

        return StringsAttributes(strings)



import os
import subprocess

import pytest

MAKEFILE_CONTENTS = """
CC=gcc

default: program

program.o: program.c $(HEADERS)
	$(CC) -c program.c -fno-asynchronous-unwind-tables -o program.o

program: program.o
	$(CC) program.o -o program

program_no_reloc: program.o
	$(CC) program.o -no-pie -o program_no_reloc

program_relocated: program_relocated.o
	$(CC) program_relocated.o -o program_relocated

large_elf.o: large_elf.c $(HEADERS)
	$(CC) -c large_elf.c -o large_elf.o

large_elf: large_elf.o
	$(CC) large_elf.o -no-pie -o large_elf 
"""

C_SOURCE_CONTENTS = """
#include <stdio.h>

int foo();
int bar();

int main() {
   printf("Hello, World!\\n");
   return foo();
}

int foo() {
    return 12;
}

int bar() {
    return 24;
}


"""

LARGE_SOURCE_CONTENTS_HEADER = """
int foo();
int bar();

int main() {
   return bar();
}

"""

LARGE_SOURCE_CONTENTS_FOOTER = """

int foo() {
    return 12;
}

int bar() {
    return 24;
}


"""

PATCH_CONTENTS = """
int noop0();

int baz()
{
    noop0();
    return 36;
}
"""


def create_noops():
    noops = []

    i = 0

    for i in range(8000):
        instruction = f"int noop{i}(){{}}"
        noops.append(instruction)

    return noops


@pytest.fixture
def elf_test_directory(tmpdir):
    makefile_path = os.path.join(tmpdir, "Makefile")
    c_source_path = os.path.join(tmpdir, "program.c")
    large_source_path = os.path.join(tmpdir, "large_elf.c")

    patch_dir = os.path.join(tmpdir, "source_dir")
    if not os.path.exists(patch_dir):
        os.mkdir(patch_dir)
    patch_path = os.path.join(patch_dir, "patch.c")

    noops = create_noops()

    LARGE_SOURCE_CONTENTS = LARGE_SOURCE_CONTENTS_HEADER
    for noop in noops:
        LARGE_SOURCE_CONTENTS += noop
    LARGE_SOURCE_CONTENTS += LARGE_SOURCE_CONTENTS_FOOTER

    with open(makefile_path, "w") as f:
        f.write(MAKEFILE_CONTENTS)
    with open(c_source_path, "w") as f:
        f.write(C_SOURCE_CONTENTS)
    with open(large_source_path, "w") as f:
        f.write(LARGE_SOURCE_CONTENTS)
    with open(patch_path, "w") as f:
        f.write(PATCH_CONTENTS)

    return tmpdir


@pytest.fixture
def elf_object_file(elf_test_directory):
    subprocess.run(["make", "-C", elf_test_directory, "program.o"])
    return os.path.join(elf_test_directory, "program.o")


@pytest.fixture
def elf_executable_file(elf_test_directory):
    subprocess.run(["make", "-C", elf_test_directory, "program"])
    return os.path.join(elf_test_directory, "program")


@pytest.fixture
def elf_no_pie_executable_file(elf_test_directory):
    subprocess.run(["make", "-C", elf_test_directory, "program_no_reloc"])
    return os.path.join(elf_test_directory, "program_no_reloc")


@pytest.fixture
def large_elf_source_file(elf_test_directory):
    return os.path.join(elf_test_directory, "large_elf.c")


@pytest.fixture
def large_elf_object_file(elf_test_directory):
    subprocess.run(["make", "-C", elf_test_directory, "large_elf.o"])
    return os.path.join(elf_test_directory, "large_elf.o")


@pytest.fixture
def large_elf_file(elf_test_directory):
    subprocess.run(["make", "-C", elf_test_directory, "large_elf"])
    return os.path.join(elf_test_directory, "large_elf")


@pytest.fixture
def patch_file(elf_test_directory):
    return os.path.join(elf_test_directory, "source_dir", "patch.c")



import re
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Iterable, Optional

from ofrak.model.viewable_tag_model import AttributesType

from ofrak_type.architecture import InstructionSet
from ofrak.core.program import Program
from ofrak.core.program_section import NamedProgramSection, ProgramSegment
from ofrak.model.resource_model import index
from ofrak.resource_view import ResourceView
from ofrak.service.resource_service_i import (
    ResourceFilter,
    ResourceAttributeValueFilter,
    ResourceSortDirection,
    ResourceSort,
)
from ofrak.core.magic import MagicDescriptionIdentifier
from ofrak_type.bit_width import BitWidth
from ofrak_type.endianness import Endianness
from ofrak_type.memory_permissions import MemoryPermissions
from ofrak_type.range import Range


SECTION_NAME_PATTERN = re.compile(b"[^\x00]*\x00")


##################################################################################
#                           ELF BASIC HEADER
##################################################################################


@dataclass
class ElfBasicHeader(ResourceView):
    """
    See "e_ident" in <https://man7.org/linux/man-pages/man5/elf.5.html> for details.
    """

    ei_magic: bytes
    ei_class: int
    ei_data: int
    ei_version: int
    ei_osabi: int
    ei_abiversion: int
    ei_pad: bytes

    def get_endianness(self) -> Endianness:
        if self.ei_data == 1:
            return Endianness.LITTLE_ENDIAN
        elif self.ei_data == 2:
            return Endianness.BIG_ENDIAN
        else:
            raise ValueError("Invalid endianness value in the ELF header ")

    def get_bitwidth(self) -> BitWidth:
        if self.ei_class == 1:
            return BitWidth.BIT_32
        elif self.ei_class == 2:
            return BitWidth.BIT_64
        else:
            raise ValueError("Invalid bit width  value in the ELF header ")


##################################################################################
#                           ELF HEADER
##################################################################################


class ElfMachine(Enum):
    EM_NONE = 0  # No machine
    EM_M32 = 1  # AT&T WE 32100
    EM_SPARC = 2  # SPARC
    EM_386 = 3  # Intel 80386
    EM_68K = 4  # Motorola 68000
    EM_88K = 5  # Motorola 88000
    EM_860 = 7  # Intel 80860
    EM_MIPS = 8  # MIPS I Architecture
    EM_S370 = 9  # IBM System/370 Processor
    EM_MIPS_RS3_LE = 10  # MIPS RS3000 Little-endian
    EM_PARISC = 15  # Hewlett-Packard PA-RISC
    EM_VPP500 = 17  # Fujitsu VPP500
    EM_SPARC32PLUS = 18  # Enhanced instruction set SPARC
    EM_960 = 19  # Intel 80960
    EM_PPC = 20  # PowerPC
    EM_PPC64 = 21  # 64-bit PowerPC
    EM_S390 = 22  # IBM System/390 Processor
    EM_V800 = 36  # NEC V800
    EM_FR20 = 37  # Fujitsu FR20
    EM_RH32 = 38  # TRW RH-32
    EM_RCE = 39  # Motorola RCE
    EM_ARM = 40  # Advanced RISC Machines ARM
    EM_ALPHA = 41  # Digital Alpha
    EM_SH = 42  # Hitachi SH
    EM_SPARCV9 = 43  # SPARC Version 9
    EM_TRICORE = 44  # Siemens TriCore embedded processor
    EM_ARC = 45  # Argonaut RISC Core, Argonaut Technologies Inc.
    EM_H8_300 = 46  # Hitachi H8/300
    EM_H8_300H = 47  # Hitachi H8/300H
    EM_H8S = 48  # Hitachi H8S
    EM_H8_500 = 49  # Hitachi H8/500
    EM_IA_64 = 50  # Intel IA-64 processor architecture
    EM_MIPS_X = 51  # Stanford MIPS-X
    EM_COLDFIRE = 52  # Motorola ColdFire
    EM_68HC12 = 53  # Motorola M68HC12
    EM_MMA = 54  # Fujitsu MMA Multimedia Accelerator
    EM_PCP = 55  # Siemens PCP
    EM_NCPU = 56  # Sony nCPU embedded RISC processor
    EM_NDR1 = 57  # Denso NDR1 microprocessor
    EM_STARCORE = 58  # Motorola Star*Core processor
    EM_ME16 = 59  # Toyota ME16 processor
    EM_ST100 = 60  # STMicroelectronics ST100 processor
    EM_TINYJ = 61  # Advanced Logic Corp. TinyJ embedded processor family
    EM_X86_64 = 62  # AMD x86-64 architecture
    EM_PDSP = 63  # Sony DSP Processor
    EM_PDP10 = 64  # Digital Equipment Corp. PDP-10
    EM_PDP11 = 65  # Digital Equipment Corp. PDP-11
    EM_FX66 = 66  # Siemens FX66 microcontroller
    EM_ST9PLUS = 67  # STMicroelectronics ST9+ 8/16 bit microcontroller
    EM_ST7 = 68  # STMicroelectronics ST7 8-bit microcontroller
    EM_68HC16 = 69  # Motorola MC68HC16 Microcontroller
    EM_68HC11 = 70  # Motorola MC68HC11 Microcontroller
    EM_68HC08 = 71  # Motorola MC68HC08 Microcontroller
    EM_68HC05 = 72  # Motorola MC68HC05 Microcontroller
    EM_SVX = 73  # Silicon Graphics SVx
    EM_ST19 = 74  # STMicroelectronics ST19 8-bit microcontroller
    EM_VAX = 75  # Digital VAX
    EM_CRIS = 76  # Axis Communications 32-bit embedded processor
    EM_JAVELIN = 77  # Infineon Technologies 32-bit embedded processor
    EM_FIREPATH = 78  # Element 14 64-bit DSP Processor
    EM_ZSP = 79  # LSI Logic 16-bit DSP Processor
    EM_MMIX = 80  # Donald Knuth's educational 64-bit processor
    EM_HUANY = 81  # Harvard University machine-independent object files
    EM_PRISM = 82  # SiTera Prism
    EM_AVR = 83  # Atmel AVR 8-bit microcontroller
    EM_FR30 = 84  # Fujitsu FR30
    EM_D10V = 85  # Mitsubishi D10V
    EM_D30V = 86  # Mitsubishi D30V
    EM_V850 = 87  # NEC v850
    EM_M32R = 88  # Mitsubishi M32R
    EM_MN10300 = 89  # Matsushita MN10300
    EM_MN10200 = 90  # Matsushita MN10200
    EM_PJ = 91  # picoJava
    EM_OPENRISC = 92  # OpenRISC 32-bit embedded processor
    EM_ARC_A5 = 93  # ARC Cores Tangent-A5
    EM_XTENSA = 94  # Tensilica Xtensa Architecture
    EM_VIDEOCORE = 95  # Alphamosaic VideoCore processor
    EM_TMM_GPP = 96  # Thompson Multimedia General Purpose Processor
    EM_NS32K = 97  # National Semiconductor 32000 series
    EM_TPC = 98  # Tenor Network TPC processor
    EM_SNP1K = 99  # Trebia SNP 1000 processor
    EM_ST200 = 100  # STMicroelectronics (www.st.com) ST200 microcontroller
    EM_MAXQ30 = 169  # Dallas Semiconductor MAXQ30 Core Micro-controllers
    EM_AARCH64 = 183  # 64-bit Advanced RISC Machines ARM

    @staticmethod
    def get_isa(e_machine: int) -> InstructionSet:
        MACHINE_TO_ISA = {
            ElfMachine.EM_386.value: InstructionSet.X86,
            ElfMachine.EM_MIPS.value: InstructionSet.MIPS,
            ElfMachine.EM_MIPS_RS3_LE.value: InstructionSet.MIPS,
            ElfMachine.EM_PPC.value: InstructionSet.PPC,
            ElfMachine.EM_PPC64.value: InstructionSet.PPC,
            ElfMachine.EM_ARM.value: InstructionSet.ARM,
            ElfMachine.EM_X86_64.value: InstructionSet.X86,
            ElfMachine.EM_AARCH64.value: InstructionSet.AARCH64,
            ElfMachine.EM_68K.value: InstructionSet.M68K,
            ElfMachine.EM_COLDFIRE.value: InstructionSet.M68K,
            ElfMachine.EM_AVR.value: InstructionSet.AVR,
            # While there is an assembler for MaxQ (
            # https://www.maximintegrated.com/content/dam/files/design/tools/tech-docs/4465
            # /AN4465-dev-tools-guide.pdf), in practice PPC is quite similar.
            ElfMachine.EM_MAXQ30.value: InstructionSet.PPC,
        }

        if e_machine not in MACHINE_TO_ISA:
            raise KeyError(
                f"ELF header e_machine value corresponds to unimplemented ISA:" f" {e_machine}"
            )
        return MACHINE_TO_ISA[e_machine]


class ElfType(Enum):
    ET_NONE = 0  # No file type
    ET_REL = 1  # Relocatable file
    ET_EXEC = 2  # Executable file
    ET_DYN = 3  # Shared object file
    ET_CORE = 4  # Core file
    ET_LOOS = 0xFE00  # Operating system-specific
    ET_HIOS = 0xFEFF  # Operating system-specific
    ET_LOPROC = 0xFF00  # Processor-specific
    ET_HIPROC = 0xFFFF  # Processor-specific


@dataclass
class ElfHeader(ResourceView):
    """
    See "ELF header (Ehdr)" in <https://man7.org/linux/man-pages/man5/elf.5.html> for details.
    """

    e_type: int
    e_machine: int
    e_version: int
    e_entry: int
    e_phoff: int
    e_shoff: int
    e_flags: int
    e_ehsize: int
    e_phentsize: int
    e_phnum: int
    e_shentsize: int
    e_shnum: int
    e_shstrndx: int

    def get_isa(self) -> InstructionSet:
        return ElfMachine.get_isa(self.e_machine)


##################################################################################
#                           ELF SEGMENT STRUCTURE
##################################################################################


@dataclass
class ElfSegmentStructure(ResourceView):
    segment_index: int

    @index
    def SegmentIndex(self) -> int:
        return self.segment_index

    async def get_header(self) -> "ElfProgramHeader":
        return await self.resource.get_only_sibling_as_view(
            ElfProgramHeader,
            ResourceFilter(
                tags=(ElfProgramHeader,),
                attribute_filters=(
                    ResourceAttributeValueFilter(
                        ElfSegmentStructure.SegmentIndex, self.segment_index
                    ),
                ),
            ),
        )


##################################################################################
#                           ELF PROGRAM HEADER
##################################################################################


class ElfProgramHeaderType(Enum):
    """
    See <https://elixir.bootlin.com/linux/latest/source/include/uapi/linux/elf.h#L25> for details.
    """

    UNKNOWN = -1
    NULL = 0
    LOAD = 1
    DYNAMIC = 2
    INTERP = 3
    NOTE = 4
    SHLIB = 5
    PHDR = 6
    TLS = 7
    GNU_EH_FRAME = 0x6474E550
    GNU_STACK = 0x6474E551
    GNU_RELRO = 0x6474E552


@dataclass
class ElfProgramHeader(ElfSegmentStructure):
    """
    See "Program header (Phdr)" in <https://man7.org/linux/man-pages/man5/elf.5.html> for details.

    """

    p_type: int
    p_offset: int
    p_vaddr: int
    p_paddr: int
    p_filesz: int
    p_memsz: int
    p_flags: int
    p_align: int

    def get_memory_permissions(self) -> MemoryPermissions:
        """
        Get the MemoryPermission for the ElfProgramHeader.
        """
        return MemoryPermissions(self.p_flags)

    @classmethod
    def caption(cls, all_attributes) -> str:
        try:
            elf_program_header = all_attributes[AttributesType[ElfProgramHeader]]
            try:
                p_type = ElfProgramHeaderType(elf_program_header.p_type).name
            except ValueError:
                p_type = f"p_type={hex(elf_program_header.p_type)}"
            return (
                f"ElfProgramHeader: {p_type}, {MemoryPermissions(elf_program_header.p_flags).name}"
            )
        except ValueError:
            return super().caption(all_attributes)


##################################################################################
#                           ELF SEGMENT
##################################################################################


@dataclass
class ElfSegment(ElfSegmentStructure, ProgramSegment):
    """
    An ELF Segment.
    """


##################################################################################
#                           ELF SECTION STRUCTURE
##################################################################################
@dataclass
class ElfSectionStructure(ResourceView):
    section_index: int

    @index
    def SectionIndex(self) -> int:
        return self.section_index

    async def get_elf(self) -> "Elf":
        return await self.resource.get_parent_as_view(Elf)

    async def get_header(self) -> "ElfSectionHeader":
        return await self.resource.get_only_sibling_as_view(
            ElfSectionHeader,
            ResourceFilter(
                tags=(ElfSectionHeader,),
                attribute_filters=(
                    ResourceAttributeValueFilter(
                        ElfSectionStructure.SectionIndex, self.section_index
                    ),
                ),
            ),
        )


##################################################################################
#                           ELF SECTION HEADER
##################################################################################


class ElfSectionType(Enum):
    UNKNOWN_OS_SPECIFIC = -0x1
    NULL = 0x0
    PROGBITS = 0x1
    SYMTAB = 0x2
    STRTAB = 0x3
    RELA = 0x4
    HASH = 0x5
    DYNAMIC = 0x6
    NOTE = 0x7
    NOBITS = 0x8
    REL = 0x9
    SHLIB = 0xA
    DYNSYM = 0xB
    INIT_ARRAY = 0xE
    FINI_ARRAY = 0xF
    PREINIT_ARRAY = 0x10
    GROUP = 0x11
    SYMTAB_SHNDX = 0x12
    NUM = 0x13
    HIPROC = 0x7FFFFFFF
    HIUSER = 0xFFFFFFFF
    LOPROC = 0x70000000
    LOUSER = 0x80000000

    GNU_HASH = 0x6FFFFFF6
    VERNEED = 0x6FFFFFFE
    VERSYM = 0x6FFFFFFF

    @classmethod
    def _missing_(cls, value):
        if value >= 0x60000000:
            return ElfSectionType.UNKNOWN_OS_SPECIFIC
        else:
            super()._missing_(value)


class ElfSectionFlag(Enum):
    WRITE = 0x1
    ALLOC = 0x2
    EXECINSTR = 0x4
    MERGE = 0x10
    STRINGS = 0x20
    INFO_LINK = 0x40
    LINK_ORDER = 0x80
    OS_NONCONFORMING = 0x100
    GROUP = 0x200
    TLS = 0x400
    MASKOS = 0x0FF00000
    ORDERED = 0x40000000
    EXCLUDE = 0x80000000
    MASKPROC = 0xF0000000


@dataclass
class ElfSectionHeader(ElfSectionStructure):
    """
    See "Section header (Shdr)" in <https://man7.org/linux/man-pages/man5/elf.5.html> for details.
    """

    sh_name: int
    sh_type: int
    sh_flags: int
    sh_addr: int
    sh_offset: int
    sh_size: int
    sh_link: int
    sh_info: int
    sh_addralign: int
    sh_entsize: int

    def has_flag(self, flag: ElfSectionFlag) -> bool:
        return self.sh_flags & flag.value > 0

    def get_flags(self) -> Iterable[ElfSectionFlag]:
        for flag in ElfSectionFlag:
            if self.has_flag(flag):
                yield flag

    def get_file_range(self) -> Range:
        return Range.from_size(self.sh_offset, self.sh_size)

    def get_type(self) -> ElfSectionType:
        return ElfSectionType(self.sh_type)


##################################################################################
#                           ELF SYMBOL
##################################################################################


class ElfSymbolBinding(Enum):
    LOCAL = 0
    GLOBAL = 1
    WEAK = 2
    LOOS = 10
    HIOS = 12
    LOPROC = 13
    HIPROC = 15


class ElfSymbolType(Enum):
    NOTYPE = 0
    OBJECT = 1
    FUNC = 2
    SECTION = 3
    FILE = 4
    COMMON = 5
    TLS = 6
    LOOS = 10
    HIOS = 12
    LOPROC = 13
    HIPROC = 15


class ElfSymbolVisibility(Enum):
    DEFAULT = 0
    INTERNAL = 1
    HIDDEN = 2
    PROTECTED = 3


@dataclass
class ElfSymbolStructure(ResourceView):
    symbol_index: int

    @index
    def SymbolIndex(self) -> int:
        return self.symbol_index


@dataclass
class ElfSymbol(ElfSymbolStructure):
    """
    See "String and symbol tables" in <https://man7.org/linux/man-pages/man5/elf.5.html> for
    details.
    """

    st_name: int
    st_value: int
    st_size: int
    st_info: int
    st_other: int
    st_shndx: int

    def get_binding(self) -> ElfSymbolBinding:
        return ElfSymbolBinding(self.st_info >> 4)

    def get_type(self) -> ElfSymbolType:
        return ElfSymbolType(self.st_info & 0xF)

    def get_visibility(self) -> ElfSymbolVisibility:
        return ElfSymbolVisibility(self.st_other & 0x3)

    def get_section_index(self) -> Optional[int]:
        if self.get_type() in (
            ElfSymbolType.OBJECT,
            ElfSymbolType.FUNC,
            ElfSymbolType.SECTION,
        ):
            return self.st_shndx
        return None

    async def get_name(self) -> str:
        elf = await self.resource.get_only_ancestor_as_view(Elf, ResourceFilter.with_tags(Elf))
        string_section = await elf.get_string_section()
        ((_, raw_symbol_name),) = await string_section.resource.search_data(
            SECTION_NAME_PATTERN, start=self.st_name, max_matches=1
        )
        return raw_symbol_name.rstrip(b"\x00").decode("ascii")

    @index
    def SymbolValue(self) -> int:
        return self.st_value


##################################################################################
#                           ELF RELA
##################################################################################


class ElfRelaInfo(Enum):
    """
    An Enum for r_info in ElfRela

    Implemented in each respective arch model.py
    """

    @staticmethod
    @abstractmethod
    def type_mask(value: int) -> int:
        raise NotImplementedError()


@dataclass
class ElfRelaEntry(ResourceView):
    """
    ElfRelaEntry describes relocation information within the program. Located in .rela.* sections.

    :var r_offset: vm offset information for each relocation entry
    :var r_info: Describes the type of relocation and sometimes the symbol related to the relocation
    :var r_addend: Describes the VM offset for each relocation itself
    """

    r_offset: int
    r_info: int
    r_addend: int


##################################################################################
#                           ELF DYNAMIC TABLE
##################################################################################


class ElfDynamicTableTag(Enum):
    DT_NULL = 0
    DT_NEEDED = 1
    DT_PLTRELSZ = 2
    DT_PLTGOT = 3
    DT_HASH = 4
    DT_STRTAB = 5
    DT_SYMTAB = 6
    DT_RELA = 7
    DT_RELASZ = 8
    DT_RELAENT = 9
    DT_STRSZ = 10
    DT_SYMENT = 11
    DT_INIT = 12
    DT_FINI = 13
    DT_SONAME = 14
    DT_RPATH = 15
    DT_SDMBOLIC = 16
    DT_REL = 17
    DT_RELSZ = 18
    DT_RELENT = 19
    DT_PLTREL = 20
    DT_DEBUG = 21
    DT_TEXTREL = 22
    DT_JMPREL = 23
    DT_INIT_ARRAY = 25
    DT_FINI_ARRAY = 26
    DT_INIT_ARRAYSZ = 27
    DT_FINI_ARRAYSZ = 28
    DT_ENCODING = 32
    OLD_DT_LOOS = 0x60000000
    DT_LOOS = 0x6000000D
    DT_HIOS = 0x6FFFF000
    DT_VALRNGLO = 0x6FFFFD00
    DT_VALRNGHI = 0x6FFFFDFF
    DT_ADDRRNGLO = 0x6FFFFE00
    DT_GNU_HASH = 0x6FFFFEF5
    DT_ADDRRNGHI = 0x6FFFFEFF
    DT_VERSYM = 0x6FFFFFF0
    DT_RELACOUNT = 0x6FFFFFF9
    DT_RELCOUNT = 0x6FFFFFFA
    DT_FLAGS_1 = 0x6FFFFFFB
    DT_VERDEF = 0x6FFFFFFC
    DT_VERDEFNUM = 0x6FFFFFFD
    DT_VERNEED = 0x6FFFFFFE
    DT_VERNEEDNUM = 0x6FFFFFFF
    OLD_DT_HIOS = 0x6FFFFFFF
    DT_LOPROC = 0x70000000
    DT_HIPROC = 0x7FFFFFFF


@dataclass
class ElfDynamicEntry(ResourceView):
    """
    ElfDynamicEntry describes a .dynamic table entry.

    https://docs.oracle.com/cd/E23824_01/html/819-0690/chapter6-42444.html

    :var d_tag: one of ElfDynamicTableTag
    :var d_un: malleable word size value that changes meaning depending on d_tag
    """

    d_tag: int
    d_un: int  # either a pointer, value, or offset depending on the tag


@dataclass
class ElfVirtualAddress(ResourceView):
    """
    Wrapper for a virtual address

    :var value: an address
    """

    value: int


##################################################################################
#                           ELF SECTION
##################################################################################


@dataclass
class ElfSection(ElfSectionStructure, NamedProgramSection):
    """
    An ELF Section.
    """


@dataclass
class ElfPointerArraySection(ElfSection):
    """
    An ELF Section that can be interpreted as an array of pointers.

    TODO: Except in the case of .init_array and .fini_array, this tag must be added by hand
        as overlapping with other tags, like CodeRegion results in Data Node inconsistencies and
        errors.
    """

    async def get_entries(self):
        await self.resource.unpack()
        return await self.resource.get_children_as_view(
            ElfVirtualAddress,
            ResourceFilter(tags=(ElfVirtualAddress,)),
        )


class ElfFiniArraySection(ElfPointerArraySection):
    pass


class ElfInitArraySection(ElfPointerArraySection):
    pass


class ElfDynamicSection(ElfSection):
    """
    The .dynamic ELF Section that appears in dynamically linked ELFs.
    """

    async def get_entries(self) -> Iterable[ElfDynamicEntry]:
        await self.resource.unpack()
        return await self.resource.get_children_as_view(
            ElfDynamicEntry,
            ResourceFilter(tags=(ElfDynamicEntry,)),
        )


class ElfRelaSection(ElfSection):
    """
    An ELF .rela.* section containing structs of type Elf{32, 64}_Rela
    """

    async def get_entries(self) -> Iterable[ElfRelaEntry]:
        await self.resource.unpack()

        return await self.resource.get_children_as_view(
            ElfRelaEntry,
            ResourceFilter(tags=(ElfRelaEntry,)),
        )


class ElfSymbolSection(ElfSection):
    """
    An ELF section containing structures of type Elf{32, 64}_Sym
    """

    async def get_symbols(self) -> Iterable[ElfSymbol]:
        await self.resource.unpack()

        return await self.resource.get_children_as_view(
            ElfSymbol,
            ResourceFilter(tags=(ElfSymbol,)),
            ResourceSort(ElfSymbol.SymbolIndex, ResourceSortDirection.ASCENDANT),
        )


class ElfDynSymbolSection(ElfSymbolSection):
    pass


class ElfStringSection(ElfSectionStructure):
    """
    A section with the STRTAB flag. There may be several of these in an ELF.
    """

    async def get_section(self) -> ElfSection:
        return await self.resource.view_as(ElfSection)


class ElfSectionNameStringSection(ElfStringSection):
    """
    A section with the STRTAB flag and named ".shstrtab". There should be at most one of these per
    ELF, and it contains only strings for the names of sections. Section headers' sh_name field
    is an index within this .shstrtab section.
    """


##################################################################################
#                           ELF
##################################################################################


@dataclass
class Elf(Program):
    """
    An Executable and Linking Format (ELF) file.

    See <https://man7.org/linux/man-pages/man5/elf.5.html> for details.
    """

    async def get_header(self) -> ElfHeader:
        return await self.resource.get_only_child_as_view(
            ElfHeader, ResourceFilter.with_tags(ElfHeader)
        )

    async def get_basic_header(self) -> ElfBasicHeader:
        return await self.resource.get_only_child_as_view(
            ElfBasicHeader, ResourceFilter.with_tags(ElfBasicHeader)
        )

    async def get_segments(self) -> Iterable[ElfSegment]:
        return await self.resource.get_children_as_view(
            ElfSegment,
            ResourceFilter(tags=(ElfSegment,)),
            ResourceSort(ElfSegmentStructure.SegmentIndex, ResourceSortDirection.ASCENDANT),
        )

    async def get_sections(self) -> Iterable[ElfSection]:
        return await self.resource.get_children_as_view(
            ElfSection,
            ResourceFilter(tags=(ElfSection,)),
            ResourceSort(ElfSectionStructure.SectionIndex, ResourceSortDirection.ASCENDANT),
        )

    async def get_section_by_index(self, index: int) -> ElfSection:
        return await self.resource.get_only_child_as_view(
            ElfSection,
            ResourceFilter(
                tags=(ElfSection,),
                attribute_filters=(
                    ResourceAttributeValueFilter(ElfSectionStructure.SectionIndex, index),
                ),
            ),
        )

    async def get_section_by_name(self, name: str) -> ElfSection:
        _ = await self.get_sections()  # Forces analyzing name of all sections
        return await self.resource.get_only_child_as_view(
            ElfSection,
            ResourceFilter(
                tags=(ElfSection,),
                attribute_filters=(ResourceAttributeValueFilter(ElfSection.SectionName, name),),
            ),
        )

    async def get_section_name_string_section(self) -> ElfSectionNameStringSection:
        return await self.resource.get_only_child_as_view(
            ElfSectionNameStringSection,
            ResourceFilter(
                tags=(ElfSectionNameStringSection,),
            ),
        )

    async def get_string_section(self) -> ElfStringSection:
        for string_section in await self.resource.get_children_as_view(
            ElfStringSection,
            ResourceFilter(
                tags=(ElfStringSection,),
            ),
        ):
            if string_section.resource.has_tag(ElfSectionNameStringSection):
                continue
            section = await string_section.get_section()
            if section.name != ".strtab":
                continue
            return string_section
        raise ValueError("Could not find string section!")

    async def get_symbol_section(self) -> ElfSymbolSection:
        return await self.resource.get_only_child_as_view(
            ElfSymbolSection,
            ResourceFilter(
                tags=(ElfSymbolSection,),
            ),
        )

    async def get_section_headers(self) -> Iterable[ElfSectionHeader]:
        return await self.resource.get_children_as_view(
            ElfSectionHeader,
            ResourceFilter(tags=(ElfSectionHeader,)),
            ResourceSort(ElfSectionStructure.SectionIndex, ResourceSortDirection.ASCENDANT),
        )

    async def get_section_header_by_index(self, index: int) -> ElfSectionHeader:
        return await self.resource.get_only_child_as_view(
            ElfSectionHeader,
            ResourceFilter(
                tags=(ElfSectionHeader,),
                attribute_filters=(
                    ResourceAttributeValueFilter(ElfSectionStructure.SectionIndex, index),
                ),
            ),
        )

    async def get_program_headers(self) -> Iterable[ElfProgramHeader]:
        return await self.resource.get_children_as_view(
            ElfProgramHeader,
            ResourceFilter(tags=(ElfProgramHeader,)),
            ResourceSort(ElfProgramHeader.SegmentIndex, ResourceSortDirection.ASCENDANT),
        )

    async def get_program_header_by_index(self, index: int) -> ElfProgramHeader:
        return await self.resource.get_only_child_as_view(
            ElfProgramHeader,
            ResourceFilter(
                tags=(ElfProgramHeader,),
                attribute_filters=(
                    ResourceAttributeValueFilter(ElfProgramHeader.SegmentIndex, index),
                ),
            ),
        )


MagicDescriptionIdentifier.register(Elf, lambda s: s.startswith("ELF "))



from typing import Tuple, Type, Any, Union, Callable, TYPE_CHECKING

from abc import ABCMeta, abstractmethod

from ofrak.service.serialization.pjson_types import PJSONType

if TYPE_CHECKING:
    from ofrak.service.serialization.pjson import PJSONSerializationService


class SerializerInterface(metaclass=ABCMeta):
    # This is set from the `SerializationService` class itself, see its __init__()
    _service: "PJSONSerializationService"

    @property
    @abstractmethod
    def targets(self) -> Tuple[Union[Type, Callable[[Type], bool]], ...]:
        """
        Mandatory attribute telling the dependency injector which types this serializer is responsible for.

        Targets can be types, or predicates of a type hint returning whether the serializer handles this type.
        """
        raise NotImplementedError()

    @abstractmethod
    def obj_to_pjson(self, obj: Any, type_hint: Any) -> PJSONType:
        """Serialize the object to PJSON. Note that the generic `self._service.to_pjson` can be used here."""
        raise NotImplementedError()

    @abstractmethod
    def pjson_to_obj(self, pjson_obj: Any, type_hint: Any) -> Any:
        """Deserialize PJSON into the object. Note that the generic `self._service.from_pjson` can be used here."""
        raise NotImplementedError()



import os
from dataclasses import dataclass
from typing import Dict, List, Union, Tuple

import pytest
from ofrak.core.filesystem import File

from ofrak.resource import Resource

from ofrak import OFRAKContext
from ofrak_type.architecture import InstructionSetMode
from ofrak.core.basic_block import BasicBlock
from ofrak.core.instruction import Instruction
from ofrak.service.resource_service_i import ResourceFilter, ResourceSort
from ofrak.core.elf.model import Elf
from pytest_ofrak.patterns import TEST_PATTERN_ASSETS_DIR
from pytest_ofrak.patterns.unpack_verify import (
    UnpackAndVerifyTestCase,
    UnpackAndVerifyPattern,
)
from ofrak_type.range import Range

ExpectedBasicBlockUnpackResult = Union[Instruction, Tuple[Instruction, ...]]


@dataclass
class BasicBlockUnpackerTestCase(
    UnpackAndVerifyTestCase[int, List[ExpectedBasicBlockUnpackResult]]
):
    binary_filename: str
    binary_md5_digest: str
    basic_block_data_ranges_in_root: Dict[int, Range]  # Used when created basic blocks manually


BASIC_BLOCK_UNPACKER_TEST_CASES = [
    BasicBlockUnpackerTestCase(
        "x64",
        {
            0x4003E0: [
                Instruction(
                    virtual_address=0x4003E0,
                    size=2,
                    disassembly="xor ebp, ebp",
                    mnemonic="xor",
                    operands="ebp, ebp",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4003E2,
                    size=3,
                    disassembly="mov r9, rdx",
                    mnemonic="mov",
                    operands="r9, rdx",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4003E5,
                    size=1,
                    disassembly="pop rsi",
                    mnemonic="pop",
                    operands="rsi",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4003E6,
                    size=3,
                    disassembly="mov rdx, rsp",
                    mnemonic="mov",
                    operands="rdx, rsp",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x4003E9,
                        size=4,
                        disassembly="and rsp, -0x10",
                        mnemonic="and",
                        operands="rsp, -0x10",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x4003E9,
                        size=4,
                        disassembly="and rsp, 0xfffffffffffffff0",
                        mnemonic="and",
                        operands="rsp, 0xfffffffffffffff0",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                Instruction(
                    virtual_address=0x4003ED,
                    size=1,
                    disassembly="push rax",
                    mnemonic="push",
                    operands="rax",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4003EE,
                    size=1,
                    disassembly="push rsp",
                    mnemonic="push",
                    operands="rsp",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4003EF,
                    size=7,
                    disassembly="mov r8, 0x4004e0",
                    mnemonic="mov",
                    operands="r8, 0x4004e0",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4003F6,
                    size=7,
                    disassembly="mov rcx, 0x4004f0",
                    mnemonic="mov",
                    operands="rcx, 0x4004f0",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4003FD,
                    size=7,
                    disassembly="mov rdi, 0x4004c4",
                    mnemonic="mov",
                    operands="rdi, 0x4004c4",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400404,
                    size=5,
                    disassembly="call 0x4003c8",
                    mnemonic="call",
                    operands="0x4003c8",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400409,
                    size=1,
                    disassembly="hlt ",
                    mnemonic="hlt",
                    operands="",
                    mode=InstructionSetMode.NONE,
                ),
            ],
            0x40040C: [
                Instruction(
                    virtual_address=0x40040C,
                    size=4,
                    disassembly="sub rsp, 0x8",
                    mnemonic="sub",
                    operands="rsp, 0x8",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x400410,
                        size=7,
                        disassembly="mov rax, qword ptr [0x600840]",
                        mnemonic="mov",
                        operands="rax, qword ptr [0x600840]",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x400410,
                        size=7,
                        disassembly="mov rax, qword ptr [rip + 0x200429]",
                        mnemonic="mov",
                        operands="rax, qword ptr [rip + 0x200429]",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                Instruction(
                    virtual_address=0x400417,
                    size=3,
                    disassembly="test rax, rax",
                    mnemonic="test",
                    operands="rax, rax",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x40041A,
                        size=2,
                        disassembly="jz 0x40041e",
                        mnemonic="jz",
                        operands="0x40041e",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x40041A,
                        size=2,
                        disassembly="je 0x40041e",
                        mnemonic="je",
                        operands="0x40041e",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
            ],
            0x40041E: [
                Instruction(
                    virtual_address=0x40041E,
                    size=4,
                    disassembly="add rsp, 0x8",
                    mnemonic="add",
                    operands="rsp, 0x8",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400422,
                    size=1,
                    disassembly="ret ",
                    mnemonic="ret",
                    operands="",
                    mode=InstructionSetMode.NONE,
                ),
            ],
            0x40041C: [
                Instruction(
                    virtual_address=0x40041C,
                    size=2,
                    disassembly="call rax",
                    mnemonic="call",
                    operands="rax",
                    mode=InstructionSetMode.NONE,
                ),
            ],
            0x400430: [
                Instruction(
                    virtual_address=0x400430,
                    size=1,
                    disassembly="push rbp",
                    mnemonic="push",
                    operands="rbp",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400431,
                    size=3,
                    disassembly="mov rbp, rsp",
                    mnemonic="mov",
                    operands="rbp, rsp",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400434,
                    size=1,
                    disassembly="push rbx",
                    mnemonic="push",
                    operands="rbx",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400435,
                    size=4,
                    disassembly="sub rsp, 0x8",
                    mnemonic="sub",
                    operands="rsp, 0x8",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x400439,
                        size=7,
                        disassembly="cmp byte ptr [0x600880], 0x0",
                        mnemonic="cmp",
                        operands="byte ptr [0x600880], 0x0",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x400439,
                        size=7,
                        disassembly="cmp byte ptr [rip + 0x200440], 0x0",
                        mnemonic="cmp",
                        operands="byte ptr [rip + 0x200440], 0x0",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                (
                    Instruction(
                        virtual_address=0x400440,
                        size=2,
                        disassembly="jnz 0x40048d",
                        mnemonic="jnz",
                        operands="0x40048d",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x400440,
                        size=2,
                        disassembly="jne 0x40048d",
                        mnemonic="jne",
                        operands="0x40048d",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
            ],
            0x40048D: [
                Instruction(
                    virtual_address=0x40048D,
                    size=4,
                    disassembly="add rsp, 0x8",
                    mnemonic="add",
                    operands="rsp, 0x8",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400491,
                    size=1,
                    disassembly="pop rbx",
                    mnemonic="pop",
                    operands="rbx",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400492,
                    size=1,
                    disassembly="leave ",
                    mnemonic="leave",
                    operands="",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400493,
                    size=1,
                    disassembly="ret ",
                    mnemonic="ret",
                    operands="",
                    mode=InstructionSetMode.NONE,
                ),
            ],
            0x400442: [
                Instruction(
                    virtual_address=0x400442,
                    size=5,
                    disassembly="mov ebx, 0x6006a0",
                    mnemonic="mov",
                    operands="ebx, 0x6006a0",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x400447,
                        size=7,
                        disassembly="mov rax, qword ptr [0x600888]",
                        mnemonic="mov",
                        operands="rax, qword ptr [0x600888]",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x400447,
                        size=7,
                        disassembly="mov rax, qword ptr [rip + 0x20043a]",
                        mnemonic="mov",
                        operands="rax, qword ptr [rip + 0x20043a]",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                Instruction(
                    virtual_address=0x40044E,
                    size=7,
                    disassembly="sub rbx, 0x600698",
                    mnemonic="sub",
                    operands="rbx, 0x600698",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400455,
                    size=4,
                    disassembly="sar rbx, 0x3",
                    mnemonic="sar",
                    operands="rbx, 0x3",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400459,
                    size=4,
                    disassembly="sub rbx, 0x1",
                    mnemonic="sub",
                    operands="rbx, 0x1",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x40045D,
                    size=3,
                    disassembly="cmp rax, rbx",
                    mnemonic="cmp",
                    operands="rax, rbx",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x400460,
                        size=2,
                        disassembly="jnc 0x400486",
                        mnemonic="jnc",
                        operands="0x400486",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x400460,
                        size=2,
                        disassembly="jae 0x400486",
                        mnemonic="jae",
                        operands="0x400486",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
            ],
            0x400486: [
                (
                    Instruction(
                        virtual_address=0x400486,
                        size=7,
                        disassembly="mov byte ptr [0x600880], 0x1",
                        mnemonic="mov",
                        operands="byte ptr [0x600880], 0x1",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x400486,
                        size=7,
                        disassembly="mov byte ptr [rip + 0x2003f3], 0x1",
                        mnemonic="mov",
                        operands="byte ptr [rip + 0x2003f3], 0x1",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
            ],
            0x400462: [
                (
                    Instruction(
                        virtual_address=0x400462,
                        size=6,
                        disassembly="nop word ptr [rax + rax*0x1]",
                        mnemonic="nop",
                        operands="word ptr [rax + rax*0x1]",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x400462,
                        size=6,
                        disassembly="nop word ptr [rax + rax]",
                        mnemonic="nop",
                        operands="word ptr [rax + rax]",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
            ],
            0x400468: [
                Instruction(
                    virtual_address=0x400468,
                    size=4,
                    disassembly="add rax, 0x1",
                    mnemonic="add",
                    operands="rax, 0x1",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x40046C,
                        size=7,
                        disassembly="mov qword ptr [0x600888], rax",
                        mnemonic="mov",
                        operands="qword ptr [0x600888], rax",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x40046C,
                        size=7,
                        disassembly="mov qword ptr [rip + 0x200415], rax",
                        mnemonic="mov",
                        operands="qword ptr [rip + 0x200415], rax",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                (
                    Instruction(
                        virtual_address=0x400473,
                        size=7,
                        disassembly="call qword ptr [0x600698 + rax*0x8]",
                        mnemonic="call",
                        operands="qword ptr [0x600698 + rax*0x8]",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x400473,
                        size=7,
                        disassembly="call qword ptr [rax*0x8 + 0x600698]",
                        mnemonic="call",
                        operands="qword ptr [rax*0x8 + 0x600698]",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                (
                    Instruction(
                        virtual_address=0x40047A,
                        size=7,
                        disassembly="mov rax, qword ptr [0x600888]",
                        mnemonic="mov",
                        operands="rax, qword ptr [0x600888]",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x40047A,
                        size=7,
                        disassembly="mov rax, qword ptr [rip + 0x200407]",
                        mnemonic="mov",
                        operands="rax, qword ptr [rip + 0x200407]",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                Instruction(
                    virtual_address=0x400481,
                    size=3,
                    disassembly="cmp rax, rbx",
                    mnemonic="cmp",
                    operands="rax, rbx",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x400484,
                        size=2,
                        disassembly="jc 0x400468",
                        mnemonic="jc",
                        operands="0x400468",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x400484,
                        size=2,
                        disassembly="jb 0x400468",
                        mnemonic="jb",
                        operands="0x400468",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
            ],
            0x4004A0: [
                (
                    Instruction(
                        virtual_address=0x4004A0,
                        size=8,
                        disassembly="cmp qword ptr [0x6006a8], 0x0",
                        mnemonic="cmp",
                        operands="qword ptr [0x6006a8], 0x0",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x4004A0,
                        size=8,
                        disassembly="cmp qword ptr [rip + 0x200200], 0x0",
                        mnemonic="cmp",
                        operands="qword ptr [rip + 0x200200], 0x0",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                Instruction(
                    virtual_address=0x4004A8,
                    size=1,
                    disassembly="push rbp",
                    mnemonic="push",
                    operands="rbp",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4004A9,
                    size=3,
                    disassembly="mov rbp, rsp",
                    mnemonic="mov",
                    operands="rbp, rsp",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x4004AC,
                        size=2,
                        disassembly="jz 0x4004c0",
                        mnemonic="jz",
                        operands="0x4004c0",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x4004AC,
                        size=2,
                        disassembly="je 0x4004c0",
                        mnemonic="je",
                        operands="0x4004c0",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
            ],
            0x4004C0: [
                Instruction(
                    virtual_address=0x4004C0,
                    size=1,
                    disassembly="leave ",
                    mnemonic="leave",
                    operands="",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4004C1,
                    size=1,
                    disassembly="ret ",
                    mnemonic="ret",
                    operands="",
                    mode=InstructionSetMode.NONE,
                ),
            ],
            0x4004AE: [
                Instruction(
                    virtual_address=0x4004AE,
                    size=5,
                    disassembly="mov eax, 0x0",
                    mnemonic="mov",
                    operands="eax, 0x0",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4004B3,
                    size=3,
                    disassembly="test rax, rax",
                    mnemonic="test",
                    operands="rax, rax",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x4004B6,
                        size=2,
                        disassembly="jz 0x4004c0",
                        mnemonic="jz",
                        operands="0x4004c0",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x4004B6,
                        size=2,
                        disassembly="je 0x4004c0",
                        mnemonic="je",
                        operands="0x4004c0",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
            ],
            0x4004B8: [
                Instruction(
                    virtual_address=0x4004B8,
                    size=5,
                    disassembly="mov edi, 0x6006a8",
                    mnemonic="mov",
                    operands="edi, 0x6006a8",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4004BD,
                    size=1,
                    disassembly="leave ",
                    mnemonic="leave",
                    operands="",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4004BE,
                    size=2,
                    disassembly="jmp rax",
                    mnemonic="jmp",
                    operands="rax",
                    mode=InstructionSetMode.NONE,
                ),
            ],
            0x4004C4: [
                Instruction(
                    virtual_address=0x4004C4,
                    size=1,
                    disassembly="push rbp",
                    mnemonic="push",
                    operands="rbp",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4004C5,
                    size=3,
                    disassembly="mov rbp, rsp",
                    mnemonic="mov",
                    operands="rbp, rsp",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4004C8,
                    size=4,
                    disassembly="sub rsp, 0x10",
                    mnemonic="sub",
                    operands="rsp, 0x10",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4004CC,
                    size=5,
                    disassembly="mov edi, 0x4005d8",
                    mnemonic="mov",
                    operands="edi, 0x4005d8",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4004D1,
                    size=5,
                    disassembly="call 0x4003b8",
                    mnemonic="call",
                    operands="0x4003b8",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4004D6,
                    size=5,
                    disassembly="mov eax, 0x0",
                    mnemonic="mov",
                    operands="eax, 0x0",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4004DB,
                    size=1,
                    disassembly="leave ",
                    mnemonic="leave",
                    operands="",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4004DC,
                    size=1,
                    disassembly="ret ",
                    mnemonic="ret",
                    operands="",
                    mode=InstructionSetMode.NONE,
                ),
            ],
            0x4004E0: [
                (
                    Instruction(
                        virtual_address=0x4004E0,
                        size=2,
                        disassembly="ret ",
                        mnemonic="ret",
                        operands="",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x4004E0,
                        size=2,
                        disassembly="rep ret ",
                        mnemonic="rep ret",
                        operands="",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
            ],
            0x4004F0: [
                Instruction(
                    virtual_address=0x4004F0,
                    size=5,
                    disassembly="mov qword ptr [rsp - 0x28], rbp",
                    mnemonic="mov",
                    operands="qword ptr [rsp - 0x28], rbp",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4004F5,
                    size=5,
                    disassembly="mov qword ptr [rsp - 0x20], r12",
                    mnemonic="mov",
                    operands="qword ptr [rsp - 0x20], r12",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x4004FA,
                        size=7,
                        disassembly="lea rbp, [0x600684]",
                        mnemonic="lea",
                        operands="rbp, [0x600684]",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x4004FA,
                        size=7,
                        disassembly="lea rbp, [rip + 0x200183]",
                        mnemonic="lea",
                        operands="rbp, [rip + 0x200183]",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                (
                    Instruction(
                        virtual_address=0x400501,
                        size=7,
                        disassembly="lea r12, [0x600684]",
                        mnemonic="lea",
                        operands="r12, [0x600684]",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x400501,
                        size=7,
                        disassembly="lea r12, [rip + 0x20017c]",
                        mnemonic="lea",
                        operands="r12, [rip + 0x20017c]",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                Instruction(
                    virtual_address=0x400508,
                    size=5,
                    disassembly="mov qword ptr [rsp - 0x18], r13",
                    mnemonic="mov",
                    operands="qword ptr [rsp - 0x18], r13",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x40050D,
                    size=5,
                    disassembly="mov qword ptr [rsp - 0x10], r14",
                    mnemonic="mov",
                    operands="qword ptr [rsp - 0x10], r14",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400512,
                    size=5,
                    disassembly="mov qword ptr [rsp - 0x8], r15",
                    mnemonic="mov",
                    operands="qword ptr [rsp - 0x8], r15",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400517,
                    size=5,
                    disassembly="mov qword ptr [rsp - 0x30], rbx",
                    mnemonic="mov",
                    operands="qword ptr [rsp - 0x30], rbx",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x40051C,
                    size=4,
                    disassembly="sub rsp, 0x38",
                    mnemonic="sub",
                    operands="rsp, 0x38",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400520,
                    size=3,
                    disassembly="sub rbp, r12",
                    mnemonic="sub",
                    operands="rbp, r12",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400523,
                    size=3,
                    disassembly="mov r13d, edi",
                    mnemonic="mov",
                    operands="r13d, edi",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400526,
                    size=3,
                    disassembly="mov r14, rsi",
                    mnemonic="mov",
                    operands="r14, rsi",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400529,
                    size=4,
                    disassembly="sar rbp, 0x3",
                    mnemonic="sar",
                    operands="rbp, 0x3",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x40052D,
                    size=3,
                    disassembly="mov r15, rdx",
                    mnemonic="mov",
                    operands="r15, rdx",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400530,
                    size=5,
                    disassembly="call 0x400390",
                    mnemonic="call",
                    operands="0x400390",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400535,
                    size=3,
                    disassembly="test rbp, rbp",
                    mnemonic="test",
                    operands="rbp, rbp",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x400538,
                        size=2,
                        disassembly="jz 0x400556",
                        mnemonic="jz",
                        operands="0x400556",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x400538,
                        size=2,
                        disassembly="je 0x400556",
                        mnemonic="je",
                        operands="0x400556",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
            ],
            0x400556: [
                Instruction(
                    virtual_address=0x400556,
                    size=5,
                    disassembly="mov rbx, qword ptr [rsp + 0x8]",
                    mnemonic="mov",
                    operands="rbx, qword ptr [rsp + 0x8]",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x40055B,
                    size=5,
                    disassembly="mov rbp, qword ptr [rsp + 0x10]",
                    mnemonic="mov",
                    operands="rbp, qword ptr [rsp + 0x10]",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400560,
                    size=5,
                    disassembly="mov r12, qword ptr [rsp + 0x18]",
                    mnemonic="mov",
                    operands="r12, qword ptr [rsp + 0x18]",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400565,
                    size=5,
                    disassembly="mov r13, qword ptr [rsp + 0x20]",
                    mnemonic="mov",
                    operands="r13, qword ptr [rsp + 0x20]",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x40056A,
                    size=5,
                    disassembly="mov r14, qword ptr [rsp + 0x28]",
                    mnemonic="mov",
                    operands="r14, qword ptr [rsp + 0x28]",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x40056F,
                    size=5,
                    disassembly="mov r15, qword ptr [rsp + 0x30]",
                    mnemonic="mov",
                    operands="r15, qword ptr [rsp + 0x30]",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400574,
                    size=4,
                    disassembly="add rsp, 0x38",
                    mnemonic="add",
                    operands="rsp, 0x38",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400578,
                    size=1,
                    disassembly="ret ",
                    mnemonic="ret",
                    operands="",
                    mode=InstructionSetMode.NONE,
                ),
            ],
            0x40053A: [
                Instruction(
                    virtual_address=0x40053A,
                    size=2,
                    disassembly="xor ebx, ebx",
                    mnemonic="xor",
                    operands="ebx, ebx",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x40053C,
                    size=4,
                    disassembly="nop dword ptr [rax]",
                    mnemonic="nop",
                    operands="dword ptr [rax]",
                    mode=InstructionSetMode.NONE,
                ),
            ],
            0x400540: [
                Instruction(
                    virtual_address=0x400540,
                    size=3,
                    disassembly="mov rdx, r15",
                    mnemonic="mov",
                    operands="rdx, r15",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400543,
                    size=3,
                    disassembly="mov rsi, r14",
                    mnemonic="mov",
                    operands="rsi, r14",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400546,
                    size=3,
                    disassembly="mov edi, r13d",
                    mnemonic="mov",
                    operands="edi, r13d",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400549,
                    size=4,
                    disassembly="call qword ptr [r12 + rbx*0x8]",
                    mnemonic="call",
                    operands="qword ptr [r12 + rbx*0x8]",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x40054D,
                        size=4,
                        disassembly="add rbx, 0x1",
                        mnemonic="add",
                        operands="rbx, 0x1",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x40054D,
                        size=4,
                        disassembly="add rbx, 1",
                        mnemonic="add",
                        operands="rbx, 1",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                Instruction(
                    virtual_address=0x400551,
                    size=3,
                    disassembly="cmp rbx, rbp",
                    mnemonic="cmp",
                    operands="rbx, rbp",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x400554,
                        size=2,
                        disassembly="jc 0x400540",
                        mnemonic="jc",
                        operands="0x400540",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x400554,
                        size=2,
                        disassembly="jb 0x400540",
                        mnemonic="jb",
                        operands="0x400540",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
            ],
            0x400580: [
                Instruction(
                    virtual_address=0x400580,
                    size=1,
                    disassembly="push rbp",
                    mnemonic="push",
                    operands="rbp",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400581,
                    size=3,
                    disassembly="mov rbp, rsp",
                    mnemonic="mov",
                    operands="rbp, rsp",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400584,
                    size=1,
                    disassembly="push rbx",
                    mnemonic="push",
                    operands="rbx",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x400585,
                    size=4,
                    disassembly="sub rsp, 0x8",
                    mnemonic="sub",
                    operands="rsp, 0x8",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x400589,
                        size=7,
                        disassembly="mov rax, qword ptr [0x600688]",
                        mnemonic="mov",
                        operands="rax, qword ptr [0x600688]",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x400589,
                        size=7,
                        disassembly="mov rax, qword ptr [rip + 0x2000f8]",
                        mnemonic="mov",
                        operands="rax, qword ptr [rip + 0x2000f8]",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                Instruction(
                    virtual_address=0x400590,
                    size=4,
                    disassembly="cmp rax, -0x1",
                    mnemonic="cmp",
                    operands="rax, -0x1",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x400594,
                        size=2,
                        disassembly="jz 0x4005af",
                        mnemonic="jz",
                        operands="0x4005af",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x400594,
                        size=2,
                        disassembly="je 0x4005af",
                        mnemonic="je",
                        operands="0x4005af",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
            ],
            0x4005AF: [
                Instruction(
                    virtual_address=0x4005AF,
                    size=4,
                    disassembly="add rsp, 0x8",
                    mnemonic="add",
                    operands="rsp, 0x8",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4005B3,
                    size=1,
                    disassembly="pop rbx",
                    mnemonic="pop",
                    operands="rbx",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4005B4,
                    size=1,
                    disassembly="leave ",
                    mnemonic="leave",
                    operands="",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4005B5,
                    size=1,
                    disassembly="ret ",
                    mnemonic="ret",
                    operands="",
                    mode=InstructionSetMode.NONE,
                ),
            ],
            0x400596: [
                Instruction(
                    virtual_address=0x400596,
                    size=5,
                    disassembly="mov ebx, 0x600688",
                    mnemonic="mov",
                    operands="ebx, 0x600688",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x40059B,
                        size=5,
                        disassembly="nop dword ptr [rax + rax*0x1]",
                        mnemonic="nop",
                        operands="dword ptr [rax + rax*0x1]",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x40059B,
                        size=5,
                        disassembly="nop dword ptr [rax + rax]",
                        mnemonic="nop",
                        operands="dword ptr [rax + rax]",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
            ],
            0x4005A0: [
                Instruction(
                    virtual_address=0x4005A0,
                    size=4,
                    disassembly="sub rbx, 0x8",
                    mnemonic="sub",
                    operands="rbx, 0x8",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4005A4,
                    size=2,
                    disassembly="call rax",
                    mnemonic="call",
                    operands="rax",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4005A6,
                    size=3,
                    disassembly="mov rax, qword ptr [rbx]",
                    mnemonic="mov",
                    operands="rax, qword ptr [rbx]",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x4005A9,
                    size=4,
                    disassembly="cmp rax, -0x1",
                    mnemonic="cmp",
                    operands="rax, -0x1",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x4005AD,
                        size=2,
                        disassembly="jnz 0x4005a0",
                        mnemonic="jnz",
                        operands="0x4005a0",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x4005AD,
                        size=2,
                        disassembly="jne 0x4005a0",
                        mnemonic="jne",
                        operands="0x4005a0",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
            ],
        },
        set(),  # No optional results
        "hello.out",
        "cc2de3c0cd2d0ded7543682c2470fcf0",
        {
            0x4003E0: Range(0x3E0, 0x40A),
            0x40040C: Range(0x40C, 0x41C),
            0x40041E: Range(0x41E, 0x423),
            0x40041C: Range(0x41C, 0x41E),
            0x400430: Range(0x430, 0x442),
            0x40048D: Range(0x48D, 0x494),
            0x400442: Range(0x442, 0x462),
            0x400486: Range(0x486, 0x48D),
            0x400462: Range(0x462, 0x468),
            0x400468: Range(0x468, 0x486),
            0x4004A0: Range(0x4A0, 0x4AE),
            0x4004C0: Range(0x4C0, 0x4C2),
            0x4004AE: Range(0x4AE, 0x4B8),
            0x4004B8: Range(0x4B8, 0x4C0),
            0x4004C4: Range(0x4C4, 0x4DD),
            0x4004E0: Range(0x4E0, 0x4E2),
            0x4004F0: Range(0x4F0, 0x53A),
            0x400556: Range(0x556, 0x579),
            0x40053A: Range(0x53A, 0x540),
            0x400540: Range(0x540, 0x556),
            0x400580: Range(0x580, 0x596),
            0x4005AF: Range(0x5AF, 0x5B6),
            0x400596: Range(0x596, 0x5A0),
            0x4005A0: Range(0x5A0, 0x5AF),
        },
    ),
    # ARM with literal pools
    BasicBlockUnpackerTestCase(
        "ARM with literal pools",
        {
            0x8018: [
                (
                    Instruction(
                        virtual_address=0x8018,
                        size=4,
                        disassembly="ldr r3, [pc, #0x10]",
                        mnemonic="ldr",
                        operands="r3, [pc, #0x10]",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x8018,
                        size=4,
                        disassembly="ldr r3, [0x8030]",
                        mnemonic="ldr",
                        operands="r3, [0x8030]",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                (
                    Instruction(
                        virtual_address=0x801C,
                        size=4,
                        disassembly="ldrb r2, [r3, #0x0]",
                        mnemonic="ldrb",
                        operands="r2, [r3, #0x0]",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x801C,
                        size=4,
                        disassembly="ldrb r2, [r3]",
                        mnemonic="ldrb",
                        operands="r2, [r3]",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                Instruction(
                    virtual_address=0x8020,
                    size=4,
                    disassembly="cmp r2, #0x0",
                    mnemonic="cmp",
                    operands="r2, #0x0",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x8024,
                    size=4,
                    disassembly="moveq r2, #0x1",
                    mnemonic="moveq",
                    operands="r2, #0x1",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x8028,
                        size=4,
                        disassembly="strbeq r2, [r3, #0x0]",
                        mnemonic="strbeq",
                        operands="r2, [r3, #0x0]",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x8028,
                        size=4,
                        disassembly="strbeq r2, [r3]",
                        mnemonic="strbeq",
                        operands="r2, [r3]",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                Instruction(
                    virtual_address=0x802C,
                    size=4,
                    disassembly="bx lr",
                    mnemonic="bx",
                    operands="lr",
                    mode=InstructionSetMode.NONE,
                ),
            ],
            0x8034: [
                (
                    Instruction(
                        virtual_address=0x8034,
                        size=4,
                        disassembly="ldr r0, [0x8060]",
                        mnemonic="ldr",
                        operands="r0, [0x8060]",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x8034,
                        size=4,
                        disassembly="ldr r0, [pc, #0x24]",
                        mnemonic="ldr",
                        operands="r0, [pc, #0x24]",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                (
                    Instruction(
                        virtual_address=0x8038,
                        size=4,
                        disassembly="stmdb sp!, {r3 lr}",
                        mnemonic="stmdb",
                        operands="sp!, {r3 lr}",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x8038,
                        size=4,
                        disassembly="push {r3, lr}",
                        mnemonic="push",
                        operands="{r3, lr}",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                (
                    Instruction(
                        virtual_address=0x803C,
                        size=4,
                        disassembly="ldr r3, [r0, #0x0]",
                        mnemonic="ldr",
                        operands="r3, [r0, #0x0]",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x803C,
                        size=4,
                        disassembly="ldr r3, [r0]",
                        mnemonic="ldr",
                        operands="r3, [r0]",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                Instruction(
                    virtual_address=0x8040,
                    size=4,
                    disassembly="cmp r3, #0x0",
                    mnemonic="cmp",
                    operands="r3, #0x0",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x8044,
                        size=4,
                        disassembly="beq 0x8058",
                        mnemonic="beq",
                        operands="0x8058",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x8044,
                        size=4,
                        disassembly="beq #0x8058",
                        mnemonic="beq",
                        operands="#0x8058",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
            ],
            0x8058: [
                (
                    Instruction(
                        virtual_address=0x8058,
                        size=4,
                        disassembly="ldmia sp!, {r3 lr}",
                        mnemonic="ldmia",
                        operands="sp!, {r3 lr}",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x8058,
                        size=4,
                        disassembly="pop {r3, lr}",
                        mnemonic="pop",
                        operands="{r3, lr}",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                Instruction(
                    virtual_address=0x805C,
                    size=4,
                    disassembly="bx lr",
                    mnemonic="bx",
                    operands="lr",
                    mode=InstructionSetMode.NONE,
                ),
            ],
            0x8048: [
                (
                    Instruction(
                        virtual_address=0x8048,
                        size=4,
                        disassembly="ldr r3, [0x8064]",
                        mnemonic="ldr",
                        operands="r3, [0x8064]",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x8048,
                        size=4,
                        disassembly="ldr r3, [pc, #0x14]",
                        mnemonic="ldr",
                        operands="r3, [pc, #0x14]",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                Instruction(
                    virtual_address=0x804C,
                    size=4,
                    disassembly="cmp r3, #0x0",
                    mnemonic="cmp",
                    operands="r3, #0x0",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x8050,
                    size=4,
                    disassembly="movne lr, pc",
                    mnemonic="movne",
                    operands="lr, pc",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x8054,
                    size=4,
                    disassembly="bxne r3",
                    mnemonic="bxne",
                    operands="r3",
                    mode=InstructionSetMode.NONE,
                ),
            ],
            0x8068: [
                Instruction(
                    virtual_address=0x8068,
                    size=4,
                    disassembly="str r11, [sp, #-0x4]!",
                    mnemonic="str",
                    operands="r11, [sp, #-0x4]!",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x806C,
                    size=4,
                    disassembly="add r11, sp, #0x0",
                    mnemonic="add",
                    operands="r11, sp, #0x0",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x8070,
                    size=4,
                    disassembly="sub sp, sp, #0xc",
                    mnemonic="sub",
                    operands="sp, sp, #0xc",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x8074,
                    size=4,
                    disassembly="str r0, [r11, #-0x8]",
                    mnemonic="str",
                    operands="r0, [r11, #-0x8]",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x8078,
                    size=4,
                    disassembly="str r1, [r11, #-0xc]",
                    mnemonic="str",
                    operands="r1, [r11, #-0xc]",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x807C,
                    size=4,
                    disassembly="mov r3, #0x0",
                    mnemonic="mov",
                    operands="r3, #0x0",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x8080,
                    size=4,
                    disassembly="mov r0, r3",
                    mnemonic="mov",
                    operands="r0, r3",
                    mode=InstructionSetMode.NONE,
                ),
                Instruction(
                    virtual_address=0x8084,
                    size=4,
                    disassembly="add sp, r11, #0x0",
                    mnemonic="add",
                    operands="sp, r11, #0x0",
                    mode=InstructionSetMode.NONE,
                ),
                (
                    Instruction(
                        virtual_address=0x8088,
                        size=4,
                        disassembly="ldr r11, [sp], #0x4",
                        mnemonic="ldr",
                        operands="r11, [sp], #0x4",
                        mode=InstructionSetMode.NONE,
                    ),
                    Instruction(
                        virtual_address=0x8088,
                        size=4,
                        disassembly="pop {r11}",
                        mnemonic="pop",
                        operands="{r11}",
                        mode=InstructionSetMode.NONE,
                    ),
                ),
                Instruction(
                    virtual_address=0x808C,
                    size=4,
                    disassembly="bx lr",
                    mnemonic="bx",
                    operands="lr",
                    mode=InstructionSetMode.NONE,
                ),
            ],
        },
        set(),  # No optional results
        "simple_arm_gcc.o.elf",
        "c79d1bea0398d7a9d0faa1ba68786f5e",
        {
            0x8018: Range(0x8018, 0x8030),
            0x8034: Range(0x8034, 0x8048),
            0x8058: Range(0x8058, 0x8060),
            0x8048: Range(0x8048, 0x8058),
            0x8068: Range(0x8068, 0x8090),
        },
    ),
]


class BasicBlockUnpackerUnpackAndVerifyPattern(UnpackAndVerifyPattern):
    """
    Test pattern which checks a BasicBlockUnpacker implementation. This pattern is ready to go
    off-the-shelf. All that is needed to use it is
     1) A subclass with the name prefixed with "Test", and
     2) That subclass should be in a pytest context where the supplied frak context will have a
     BasicBlockUnpacker implementation.

    This file includes test cases of the type BasicBlockUnpacker. These include a URL
    which points to a binary which serves as the root resource. This binary will be unpacked
    recursively down to Instructions, and all ComplexBlocks in the .text section will be
    extracted. Those ComplexBlocks should line up with the expected complex block virtual
    addresses; so should all of the BasicBlocks and DataWords that make up that ComplexBlock; and
    so should all of the Instructions in each of those BasicBlocks.
    Each of those are checked for some other specific expected attributes as well.
    """

    async def unpack(self, root_resource: Resource):
        await root_resource.unpack_recursively()

    @pytest.fixture(params=BASIC_BLOCK_UNPACKER_TEST_CASES, ids=lambda tc: tc.label)
    async def unpack_verify_test_case(self, request) -> BasicBlockUnpackerTestCase:
        return request.param

    @pytest.fixture
    async def root_resource(
        self,
        unpack_verify_test_case: BasicBlockUnpackerTestCase,
        ofrak_context: OFRAKContext,
        test_id: str,
    ) -> Resource:
        asset_path = os.path.join(TEST_PATTERN_ASSETS_DIR, unpack_verify_test_case.binary_filename)
        with open(asset_path, "rb") as f:
            binary_data = f.read()
        resource = await ofrak_context.create_root_resource(test_id, binary_data, tags=(File,))
        return resource

    async def verify_descendant(
        self, basic_block: BasicBlock, specified_result: List[ExpectedBasicBlockUnpackResult]
    ):
        instructions = await basic_block.get_instructions()

        # Check that the parent complex blocks are extracted as expected
        instructions_by_addr: Dict[int, ExpectedBasicBlockUnpackResult] = dict()
        for expected_instructions in specified_result:
            if type(expected_instructions) is tuple:
                instructions_by_addr[
                    expected_instructions[0].virtual_address
                ] = expected_instructions
            else:
                instructions_by_addr[expected_instructions.virtual_address] = (
                    expected_instructions,
                )

        # Check that all expected basic blocks have been extracted
        assert {instruction.virtual_address for instruction in instructions} == set(
            instructions_by_addr.keys()
        )

        errors = []
        for instruction in instructions:
            expected_instructions = instructions_by_addr[instruction.virtual_address]
            # GhidraBasicBlockUnpacker does not provide trustworthy results for
            #  - Instruction.registers_read
            #  - Instruction.registers_written
            sanitized_instruction = Instruction(
                instruction.virtual_address,
                instruction.size,
                instruction.disassembly,
                instruction.mnemonic,
                instruction.operands,
                instruction.mode,
            )
            sanitized_expected_instructions = tuple(
                Instruction(
                    expected_instr.virtual_address,
                    expected_instr.size,
                    expected_instr.disassembly,
                    expected_instr.mnemonic,
                    expected_instr.operands,
                    expected_instr.mode,
                )
                for expected_instr in expected_instructions
            )
            if not any(
                sanitized_instruction == sanitized_expected_instr
                for sanitized_expected_instr in sanitized_expected_instructions
            ):
                e = ValueError(
                    f"Extracted instruction at {instruction.virtual_address:x} does not match any "
                    f"of the possible expected results: \n"
                    f"got: {sanitized_instruction}\n"
                    f"expected any of: {sanitized_expected_instructions}"
                )
                errors.append(e)

        if len(errors) > 0:
            raise ValueError(*errors)

    async def get_descendants_to_verify(self, unpacked_resource: Resource) -> Dict[int, Resource]:
        elf = await unpacked_resource.view_as(Elf)
        text_section = await elf.get_section_by_name(".text")
        basic_blocks: List[BasicBlock] = list(
            await text_section.resource.get_descendants_as_view(
                BasicBlock,
                r_filter=ResourceFilter.with_tags(BasicBlock),
                r_sort=ResourceSort(BasicBlock.VirtualAddress),
            )
        )
        return {bb.virtual_address: bb for bb in basic_blocks}



import io
from abc import abstractmethod, ABC
from dataclasses import dataclass
from typing import Any, Dict, Iterable, Optional, Tuple, Union

from ofrak.component.modifier import Modifier
from ofrak.model.component_model import ComponentConfig
from ofrak.model.resource_model import ResourceAttributes
from ofrak.model.viewable_tag_model import AttributesType
from ofrak.resource import Resource
from ofrak.service.resource_service_i import ResourceFilter
from ofrak.core.elf.model import (
    ElfSectionHeader,
    Elf,
    ElfProgramHeader,
    ElfHeader,
    ElfSymbol,
    ElfSectionFlag,
    ElfRelaEntry,
    ElfDynamicEntry,
    ElfPointerArraySection,
    ElfVirtualAddress,
)
from ofrak_io.deserializer import BinaryDeserializer
from ofrak_io.serializer import BinarySerializer
from ofrak_type.range import Range


class AbstractElfAttributeModifier(ABC):
    @classmethod
    @abstractmethod
    def populate_serializer(cls, serializer: BinarySerializer, attributes: Any):
        raise NotImplementedError()

    async def serialize(self, elf_resource: Elf, updated_attributes: ResourceAttributes) -> bytes:
        e_basic_header_r = await elf_resource.get_basic_header()
        buf = io.BytesIO()
        serializer = BinarySerializer(
            buf,
            endianness=e_basic_header_r.get_endianness(),
            word_size=e_basic_header_r.get_bitwidth().get_word_size(),
        )
        self.populate_serializer(serializer, updated_attributes)
        return buf.getvalue()

    async def serialize_and_patch(
        self,
        resource: Resource,
        original_attributes: Any,
        modifier_config: ComponentConfig,
    ):
        elf_resource = await resource.get_only_ancestor_as_view(Elf, ResourceFilter.with_tags(Elf))
        new_attributes = ResourceAttributes.replace_updated(original_attributes, modifier_config)
        new_data = await self.serialize(elf_resource, new_attributes)
        patch_length = await resource.get_data_length()
        resource.queue_patch(Range.from_size(0, patch_length), new_data)
        resource.add_attributes(new_attributes)


@dataclass
class ElfHeaderModifierConfig(ComponentConfig):
    e_type: Optional[int] = None
    e_machine: Optional[int] = None
    e_version: Optional[int] = None
    e_entry: Optional[int] = None
    e_phoff: Optional[int] = None
    e_shoff: Optional[int] = None
    e_flags: Optional[int] = None
    e_ehsize: Optional[int] = None
    e_phentsize: Optional[int] = None
    e_phnum: Optional[int] = None
    e_shentsize: Optional[int] = None
    e_shnum: Optional[int] = None
    e_shstrndx: Optional[int] = None


class ElfHeaderModifier(Modifier[ElfHeaderModifierConfig], AbstractElfAttributeModifier):
    id = b"ElfHeaderModifier"
    targets = (ElfHeader,)

    @classmethod
    def populate_serializer(
        cls, serializer: BinarySerializer, attributes: AttributesType[ElfHeader]
    ):
        serializer.pack_multiple(
            "HHIQQQIHHHHHH",
            attributes.e_type,
            attributes.e_machine,
            attributes.e_version,
            attributes.e_entry,
            attributes.e_phoff,
            attributes.e_shoff,
            attributes.e_flags,
            attributes.e_ehsize,
            attributes.e_phentsize,
            attributes.e_phnum,
            attributes.e_shentsize,
            attributes.e_shnum,
            attributes.e_shstrndx,
            auto_bitwidth=True,
        )

    async def modify(self, resource: Resource, config: ElfHeaderModifierConfig):
        original_attributes = await resource.analyze(AttributesType[ElfHeader])
        await self.serialize_and_patch(resource, original_attributes, config)


@dataclass
class ElfProgramHeaderModifierConfig(ComponentConfig):
    p_type: Optional[int] = None
    p_offset: Optional[int] = None
    p_vaddr: Optional[int] = None
    p_paddr: Optional[int] = None
    p_filesz: Optional[int] = None
    p_memsz: Optional[int] = None
    p_flags: Optional[int] = None
    p_align: Optional[int] = None


class ElfProgramHeaderModifier(
    AbstractElfAttributeModifier, Modifier[ElfProgramHeaderModifierConfig]
):
    targets = (ElfProgramHeader,)

    async def modify(self, resource: Resource, config: ElfProgramHeaderModifierConfig):
        original_attributes = await resource.analyze(AttributesType[ElfProgramHeader])
        await self.serialize_and_patch(resource, original_attributes, config)

    @classmethod
    def populate_serializer(
        cls, serializer: BinarySerializer, elf_program_header: ElfProgramHeader
    ):
        serializer.pack_uint(elf_program_header.p_type)
        if serializer.get_word_size() == 8:
            serializer.pack_uint(elf_program_header.p_flags)
        serializer.pack_multiple(
            "QQQQQ",
            elf_program_header.p_offset,
            elf_program_header.p_vaddr,
            elf_program_header.p_paddr,
            elf_program_header.p_filesz,
            elf_program_header.p_memsz,
            auto_bitwidth=True,
        )
        if serializer.get_word_size() == 4:
            serializer.pack_uint(elf_program_header.p_flags)
        serializer.pack_ulong(elf_program_header.p_align)


@dataclass
class ElfSectionHeaderModifierConfig(ComponentConfig):
    sh_name: Optional[int] = None
    sh_type: Optional[int] = None
    sh_flags: Optional[int] = None
    sh_addr: Optional[int] = None
    sh_offset: Optional[int] = None
    sh_size: Optional[int] = None
    sh_link: Optional[int] = None
    sh_info: Optional[int] = None
    sh_addralign: Optional[int] = None
    sh_entsize: Optional[int] = None


class ElfSectionHeaderModifier(
    AbstractElfAttributeModifier, Modifier[ElfSectionHeaderModifierConfig]
):
    id = b"ElfSectionHeaderModifier"
    targets = (ElfSectionHeader,)

    @classmethod
    def populate_serializer(
        cls,
        serializer: BinarySerializer,
        attributes: AttributesType[ElfSectionHeader],
    ):
        serializer.pack_multiple(
            "IIQQQQIIQQ",
            attributes.sh_name,
            attributes.sh_type,
            attributes.sh_flags,
            attributes.sh_addr,
            attributes.sh_offset,
            attributes.sh_size,
            attributes.sh_link,
            attributes.sh_info,
            attributes.sh_addralign,
            attributes.sh_entsize,
            auto_bitwidth=True,
        )

    async def modify(
        self,
        resource: Resource,
        config: ElfSectionHeaderModifierConfig,
    ):
        original_attributes = await resource.analyze(AttributesType[ElfSectionHeader])
        await self.serialize_and_patch(resource, original_attributes, config)


@dataclass
class ElfSymbolModifierConfig(ComponentConfig):
    st_name: Optional[int] = None
    st_value: Optional[int] = None
    st_size: Optional[int] = None
    st_info: Optional[int] = None
    st_other: Optional[int] = None
    st_shndx: Optional[int] = None


class ElfSymbolModifier(AbstractElfAttributeModifier, Modifier[ElfSymbolModifierConfig]):
    id = b"ElfSymbolModifier"
    targets = (ElfSymbol,)

    @classmethod
    def populate_serializer(
        cls, serializer: BinarySerializer, attributes: AttributesType[ElfSymbol]
    ):
        if serializer.get_word_size() == 8:
            serializer.pack_multiple(
                "IBBHQQ",
                attributes.st_name,
                attributes.st_info,
                attributes.st_other,
                attributes.st_shndx,
                attributes.st_value,
                attributes.st_size,
            )
        else:
            serializer.pack_multiple(
                "IIIBBH",
                attributes.st_name,
                attributes.st_value,
                attributes.st_size,
                attributes.st_info,
                attributes.st_other,
                attributes.st_shndx,
            )

    async def modify(
        self,
        resource: Resource,
        config: ElfSymbolModifierConfig,
    ):
        original_attributes = await resource.analyze(AttributesType[ElfSymbol])
        await self.serialize_and_patch(resource, original_attributes, config)


@dataclass
class ElfRelaModifierConfig(ComponentConfig):
    """
    :var r_offset: vm offset information for each relocation entry
    :var r_info: Describes the type of relocation and sometimes the symbol related to the relocation
    :var r_addend: Describes the VM offset for each relocation itself
    """

    r_offset: Optional[int] = None
    r_info: Optional[int] = None
    r_addend: Optional[int] = None


class ElfRelaModifier(AbstractElfAttributeModifier, Modifier[ElfRelaModifierConfig]):
    """
    The ElfRelaModifier updates values in an Elf{32, 64}_Rela struct
    """

    targets = (ElfRelaEntry,)

    @classmethod
    def populate_serializer(
        cls, serializer: BinarySerializer, attributes: AttributesType[ElfRelaEntry]
    ):
        if serializer.get_word_size() == 8:
            serializer.pack_multiple(
                "QQq",
                attributes.r_offset,
                attributes.r_info,
                attributes.r_addend,
            )
        else:
            serializer.pack_multiple(
                "IIi",
                attributes.r_offset,
                attributes.r_info,
                attributes.r_addend,
            )

    async def modify(
        self,
        resource: Resource,
        config: ElfRelaModifierConfig,
    ):
        """
        Patches the Elf{32, 64}_Rela struct
        """
        original_attributes = await resource.analyze(AttributesType[ElfRelaEntry])
        await self.serialize_and_patch(resource, original_attributes, config)


@dataclass
class ElfDynamicEntryModifierConfig(ComponentConfig):
    """
    :var d_tag: one of ElfDynamicTableTag
    :var d_un: malleable word size value that changes meaning depending on d_tag
    """

    d_tag: Optional[int] = None
    d_un: Optional[int] = None


class ElfDynamicEntryModifier(
    AbstractElfAttributeModifier, Modifier[ElfDynamicEntryModifierConfig]
):
    """
    The ElfRelaModifier updates values in an Elf{32, 64}_Dyn struct
    """

    targets = (ElfDynamicEntry,)

    @classmethod
    def populate_serializer(
        cls,
        serializer: BinarySerializer,
        attributes: AttributesType[ElfDynamicEntry],
    ):
        if serializer.get_word_size() == 8:
            serializer.pack_multiple(
                "QQ",
                attributes.d_tag,
                attributes.d_un,
            )
        else:
            serializer.pack_multiple(
                "II",
                attributes.d_tag,
                attributes.d_un,
            )

    async def modify(
        self,
        resource: Resource,
        config: ElfDynamicEntryModifierConfig,
    ):
        """
        Patches the Elf{32, 64}_Dyn struct
        """
        original_attributes = await resource.analyze(AttributesType[ElfDynamicEntry])
        await self.serialize_and_patch(resource, original_attributes, config)


@dataclass
class ElfVirtualAddressModifierConfig(ComponentConfig):
    """
    :var value: an address
    """

    value: Optional[int] = None


class ElfVirtualAddressModifier(
    AbstractElfAttributeModifier, Modifier[ElfVirtualAddressModifierConfig]
):
    """
    The ElfVirtualAddressModifier updates a pointer value
    """

    targets = (ElfVirtualAddress,)

    @classmethod
    def populate_serializer(
        cls,
        serializer: BinarySerializer,
        attributes: AttributesType[ElfVirtualAddress],
    ):
        serializer.pack_ulong(attributes.value)

    async def modify(
        self,
        resource: Resource,
        config: ElfVirtualAddressModifierConfig,
    ):
        """
        Patches the virtual address
        """
        original_attributes = await resource.analyze(AttributesType[ElfVirtualAddress])
        await self.serialize_and_patch(resource, original_attributes, config)


@dataclass
class ElfPointerArraySectionAddModifierConfig(ComponentConfig):
    """
    :var skip_list: values that should not be modified
    :var add_value: value to add to all pointers
    """

    skip_list: Iterable[int]
    add_value: int


class ElfPointerArraySectionAddModifier(Modifier[ElfPointerArraySectionAddModifierConfig]):
    """
    The ElfPointerArrayAddModifier updates batches of pointer values
    """

    targets = (ElfPointerArraySection,)

    async def modify(
        self,
        resource: Resource,
        config: ElfPointerArraySectionAddModifierConfig,
    ):
        """
        Patches the virtual addresses, doesn't change the ElfPointerArraySection attributes
        """

        elf_resource = await resource.get_only_ancestor_as_view(Elf, ResourceFilter.with_tags(Elf))
        e_basic_header_r = await elf_resource.get_basic_header()
        values = list()
        deserializer = BinaryDeserializer(
            io.BytesIO(await resource.get_data()),
            endianness=e_basic_header_r.get_endianness(),
            word_size=e_basic_header_r.get_bitwidth().get_word_size(),
        )
        num_values = (
            await resource.get_data_length() // e_basic_header_r.get_bitwidth().get_word_size()
        )
        for i in range(num_values):
            values.append(deserializer.unpack_ulong())

        buf = io.BytesIO()
        serializer = BinarySerializer(
            buf,
            endianness=e_basic_header_r.get_endianness(),
            word_size=e_basic_header_r.get_bitwidth().get_word_size(),
        )
        for value in values:
            if value not in config.skip_list:
                serializer.pack_ulong(value + config.add_value)
            else:
                serializer.pack_ulong(value)

        patch_length = await resource.get_data_length()
        resource.queue_patch(Range.from_size(0, patch_length), buf.getvalue())


@dataclass
class ElfAddStringModifierConfig(ComponentConfig):
    strings: Union[Tuple[str, ...], str]


class ElfAddStringModifier(Modifier[ElfAddStringModifierConfig]):
    """
    Adds one or more strings to the `.strtab` section in an ELF so that they can be used as the
    names for things like symbols. This modifier only inserts the strings and fixes up the
    succeeding section offsets; it does not modify any existing strings nor does it replace any
    existing strings.
    """

    targets = (Elf,)

    async def modify(
        self,
        resource: Resource,
        config: ElfAddStringModifierConfig,
    ):
        elf = await resource.view_as(Elf)
        string_section = await elf.get_string_section()
        string_section_size = await string_section.resource.get_data_length()

        if type(config.strings) is str:
            strings: Union[Tuple[str, ...], str] = (config.strings,)
        else:
            strings = config.strings
        encoded_strings = b"\x00"
        for string in strings:
            encoded_strings += string.encode("ascii") + b"\x00"
        total_string_section_size_increase = len(encoded_strings) - 1
        # Overwrites the last null byte, but our patch starts with a null byte
        string_section.resource.queue_patch(
            Range.from_size(string_section_size - 1, 1), encoded_strings
        )
        string_section_header = await string_section.get_header()
        original_string_section_offset = string_section_header.sh_offset

        # Now shift all the sections after the string section
        sections = await elf.get_sections()
        for section in sections:
            section_header = await section.get_header()
            if section_header.sh_offset <= original_string_section_offset:
                continue
            if ElfSectionFlag.ALLOC in section_header.get_flags():
                raise NotImplementedError(
                    "Expanding string section would shift offset of section "
                    "which is loaded into memory! May be possible to "
                    "handle, but this is not implemented."
                )
            await section_header.resource.run(
                ElfSectionHeaderModifier,
                ElfSectionHeaderModifierConfig(
                    sh_offset=section_header.sh_offset + total_string_section_size_increase
                ),
            )

        await string_section_header.resource.run(
            ElfSectionHeaderModifier,
            ElfSectionHeaderModifierConfig(
                sh_size=string_section_size + total_string_section_size_increase
            ),
        )

        # Section table is probably at end of binary too
        elf_header = await elf.get_header()
        if elf_header.e_shoff > string_section_header.sh_offset:
            await elf_header.resource.run(
                ElfHeaderModifier,
                ElfHeaderModifierConfig(
                    e_shoff=elf_header.e_shoff + total_string_section_size_increase
                ),
            )


@dataclass
class ElfRelocateSymbolsModifierConfig(ComponentConfig):
    new_symbol_vaddrs: Dict[int, int]


class ElfRelocateSymbolsModifier(Modifier[ElfRelocateSymbolsModifierConfig]):
    """
    Changes the virtual address value of symbols in an ELF file. If that ELF is an object file
    and is subsequently linked into an executable, any instructions referencing that symbol will
    now refer to the new address. This works even on implicitly or automatically generated symbols,
    like the absolute branches between basic blocks within a function. A linker script cannot
    change the targets of these branches individually, but this modifier can.

    The config includes a dictionary, which should map from the original address of a symbol to
    the new address that symbol should be defined as. For example, if a branch jumps to `0x1000`
    and the goal is to change that branch to instead jump to `0x1800`, the config dictionary should
    include the pair `{0x1000: 0x1800}`.
    """

    targets = (Elf,)

    async def modify(self, resource: Resource, config: ElfRelocateSymbolsModifierConfig) -> None:
        elf = await resource.view_as(Elf)
        symbol_section = await elf.get_symbol_section()
        for symbol in await symbol_section.get_symbols():
            if symbol.st_value in config.new_symbol_vaddrs:
                await symbol.resource.run(
                    ElfSymbolModifier,
                    ElfSymbolModifierConfig(
                        st_value=config.new_symbol_vaddrs[symbol.st_value],
                    ),
                )



import logging
from argparse import ArgumentDefaultsHelpFormatter, Namespace

from ofrak.cli.ofrak_cli import OfrakCommandRunsScript
from ofrak.gui.server import open_gui
from ofrak.ofrak_context import OFRAKContext

LOGGER = logging.getLogger(__name__)


class GUICommand(OfrakCommandRunsScript):
    def create_parser(self, ofrak_subparsers):
        gui_parser = ofrak_subparsers.add_parser(
            "gui",
            help="Launch the OFRAK GUI server.",
            description="Launch the OFRAK GUI server.",
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        gui_parser.add_argument(
            "-H",
            "--hostname",
            action="store",
            help="Set GUI server host address.",
            default="127.0.0.1",
        )
        gui_parser.add_argument(
            "-p",
            "--port",
            action="store",
            type=int,
            help="Set GUI server host port.",
            default=8080,
        )
        gui_parser.add_argument(
            "--no-browser",
            action="store_true",
            help="Don't open the browser to the OFRAK GUI",
        )
        gui_parser.add_argument(
            "--file",
            "-f",
            required=False,
            action="append",
            help="Path to a file to load into OFRAK when starting the GUI (multiple may be "
            "provided)",
            default=[],
        )
        gui_parser.add_argument(
            "--enable-cors",
            action="store_true",
            help="Enable CORS for debugging.",
        )
        self.add_ofrak_arguments(gui_parser)
        return gui_parser

    async def ofrak_func(self, ofrak_context: OFRAKContext, args: Namespace):  # pragma: no cover
        most_recent_root = None
        if len(args.file) > 0:
            for path in args.file:
                most_recent_root = await ofrak_context.create_root_resource_from_file(path)
        server = await open_gui(
            args.hostname,
            args.port,
            focus_resource=most_recent_root if len(args.file) == 1 else None,
            ofrak_context=ofrak_context,
            open_in_browser=(not args.no_browser),
            enable_cors=(args.enable_cors),
        )
        await server.run_until_cancelled()



import asyncio
import ctypes
import logging
import math
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures.process import BrokenProcessPool
from dataclasses import dataclass

from ofrak.component.analyzer import Analyzer
from ofrak.model.resource_model import ResourceAttributes
from ofrak.resource import Resource, ResourceFactory
from ofrak.service.data_service_i import DataServiceInterface
from ofrak.service.resource_service_i import ResourceServiceInterface

LOGGER = logging.getLogger(__name__)


C_LOG_TYPE = ctypes.CFUNCTYPE(None, ctypes.c_uint8)


try:
    from ofrak.core.entropy.entropy_c import entropy_c as entropy_func
except:
    from ofrak.core.entropy.entropy_py import entropy_py as entropy_func


@dataclass(**ResourceAttributes.DATACLASS_PARAMS)
class DataSummary(ResourceAttributes):
    """
    High-level summary of binary data.

    :ivar entropy_samples: Shannon entropy of the data. A description of Shannon entropy and how it
    can be used is [here](../../../../user-guide/key-concepts/gui/minimap.md#entropy-view).
    :ivar magnitude_samples: Sample of the binary data to put an upper limit on the displayed byte
    magnitudes; if the input data is smaller than this upper limit, all bytes are sampled.
    """

    entropy_samples: bytes
    magnitude_samples: bytes


class DataSummaryAnalyzer(Analyzer[None, DataSummary]):
    """
    Analyze binary data and return summaries of its structure via the entropy and magnitude of
    its bytes.
    """

    targets = ()  # Target any resource with data
    outputs = (DataSummary,)

    def __init__(
        self,
        resource_factory: ResourceFactory,
        data_service: DataServiceInterface,
        resource_service: ResourceServiceInterface,
    ):
        super().__init__(resource_factory, data_service, resource_service)
        self.pool = ProcessPoolExecutor()
        self.max_analysis_retries = 10

    async def analyze(self, resource: Resource, config=None, depth=0) -> DataSummary:
        if depth > self.max_analysis_retries:
            raise RuntimeError(
                f"Analysis process killed more than {self.max_analysis_retries} times. Aborting."
            )

        data = await resource.get_data()
        # Run blocking computations in separate processes
        try:
            entropy = await asyncio.get_running_loop().run_in_executor(
                self.pool, sample_entropy, data, resource.get_id()
            )
            magnitude = await asyncio.get_running_loop().run_in_executor(
                self.pool, sample_magnitude, data
            )
            return DataSummary(entropy, magnitude)
        except BrokenProcessPool:
            # If the previous one was aborted, try again with a new pool
            self.pool = ProcessPoolExecutor()
            return await self.analyze(resource, config=config, depth=depth + 1)


def sample_entropy(
    data: bytes, resource_id: bytes, window_size=256, max_samples=2**20
) -> bytes:  # pragma: no cover
    """
    Return a list of entropy values where each value represents the Shannon entropy of the byte
    value distribution over a fixed-size, sliding window. If the entropy data is larger than a
    maximum size, summarize it by periodically sampling it.

    Shannon entropy represents how uniform a probability distribution is. Since more uniform
    implies less predictable (because the probability of any outcome is equally likely in a
    uniform distribution), a sample with higher entropy is "more random" than one with lower
    entropy. More here: <https://en.wikipedia.org/wiki/Entropy_(information_theory)>.
    """

    if len(data) < 256:
        return b""

    def log_percent(percent):  # pragma: no cover
        LOGGER.info(f"Entropy calculation {percent}% complete for {resource_id.hex()}")

    result = entropy_func(data, window_size, log_percent)

    if len(result) <= max_samples:
        return result

    # Sample the calculated array if it is too large
    skip = len(result) / max_samples
    return bytes(result[math.floor(i * skip)] for i in range(max_samples))


def sample_magnitude(data: bytes, max_samples=2**20) -> bytes:  # pragma: no cover
    if len(data) < max_samples:
        # TODO: Should this be a shallow copy instead?
        return data
    else:
        skip = len(data) / max_samples
        return bytes(data[math.floor(i * skip)] for i in range(max_samples))



import asyncio
import logging
import tempfile
from gzip import GzipFile
from io import BytesIO
from subprocess import CalledProcessError

from ofrak.component.packer import Packer
from ofrak.component.unpacker import Unpacker
from ofrak.core.binary import GenericBinary
from ofrak.core.magic import MagicMimeIdentifier, MagicDescriptionIdentifier
from ofrak.model.component_model import ComponentExternalTool
from ofrak.resource import Resource
from ofrak_type.range import Range

LOGGER = logging.getLogger(__name__)

PIGZ = ComponentExternalTool(
    "pigz", "https://zlib.net/pigz/", "--help", apt_package="pigz", brew_package="pigz"
)


class GzipData(GenericBinary):
    """
    A gzip binary blob.
    """

    async def get_file(self) -> Resource:
        return await self.resource.get_only_child()


class GzipUnpacker(Unpacker[None]):
    """
    Unpack (decompress) a gzip file.
    """

    id = b"GzipUnpacker"
    targets = (GzipData,)
    children = (GenericBinary,)
    external_dependencies = (PIGZ,)

    async def unpack(self, resource: Resource, config=None):
        # Create temporary file with .gz extension
        with tempfile.NamedTemporaryFile(suffix=".gz") as temp_file:
            temp_file.write(await resource.get_data())
            temp_file.flush()
            cmd = [
                "pigz",
                "-d",
                "-c",
                temp_file.name,
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, stderr = await proc.communicate()
            data = stdout
            if proc.returncode:
                # Forward any gzip warning message and continue
                if proc.returncode == -2 or proc.returncode == 2:
                    LOGGER.warning(stderr)
                    data = stdout
                else:
                    raise CalledProcessError(returncode=proc.returncode, cmd=cmd)

            await resource.create_child(
                tags=(GenericBinary,),
                data=data,
            )


class GzipPacker(Packer[None]):
    """
    Pack data into a compressed gzip file.
    """

    targets = (GzipData,)
    external_dependencies = (PIGZ,)

    async def pack(self, resource: Resource, config=None):
        gzip_view = await resource.view_as(GzipData)

        result = BytesIO()
        with GzipFile(fileobj=result, mode="w") as gzip_file:
            gzip_child_r = await gzip_view.get_file()
            gzip_data = await gzip_child_r.get_data()
            gzip_file.write(gzip_data)

        original_gzip_size = await gzip_view.resource.get_data_length()
        resource.queue_patch(Range(0, original_gzip_size), result.getvalue())


MagicMimeIdentifier.register(GzipData, "application/gzip")
MagicDescriptionIdentifier.register(GzipData, lambda s: s.startswith("gzip compressed data"))



from pytest_ofrak.mock_component_types import MockAnalyzer
from pytest_ofrak.mock_library import (
    ITargetsRROutputsD,
    ITargetsPOutputsA,
)


class AlternativeTargetsRROutputsD(MockAnalyzer, ITargetsRROutputsD):
    pass


class AlternativeTargetsPOutputsA(MockAnalyzer, ITargetsPOutputsA):
    pass



import asyncio
import tempfile
from dataclasses import dataclass
from subprocess import CalledProcessError

from ofrak.component.unpacker import Unpacker
from ofrak.core.binary import GenericBinary
from ofrak.core.filesystem import FilesystemRoot, File, Folder, SpecialFileType

from ofrak.core.magic import MagicMimeIdentifier, MagicDescriptionIdentifier
from ofrak.model.component_model import ComponentExternalTool
from ofrak.resource import Resource
from ofrak.model.component_model import ComponentConfig

UNAR = ComponentExternalTool(
    "unar",
    "https://theunarchiver.com/command-line",
    "--help",
    apt_package="unar",
    brew_package="unar",
)


@dataclass
class RarArchive(GenericBinary, FilesystemRoot):
    """
    Filesystem stored in a RAR archive.
    """


class RarUnpacker(Unpacker[None]):
    """
    Unpack RAR archives using the free `unrar` tool.
    """

    targets = (RarArchive,)
    children = (File, Folder, SpecialFileType)
    external_dependencies = (UNAR,)

    async def unpack(self, resource: Resource, config: ComponentConfig = None):
        with tempfile.NamedTemporaryFile(
            suffix=".rar"
        ) as temp_archive, tempfile.TemporaryDirectory() as temp_dir:
            temp_archive.write(await resource.get_data())
            temp_archive.flush()

            cmd = [
                "unar",
                "-no-directory",
                "-no-recursion",
                temp_archive.name,
            ]
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                cwd=temp_dir,
            )
            returncode = await proc.wait()
            if proc.returncode:
                raise CalledProcessError(returncode=returncode, cmd=cmd)

            rar_view = await resource.view_as(RarArchive)
            await rar_view.initialize_from_disk(temp_dir)


MagicMimeIdentifier.register(RarArchive, "application/x-rar-compressed")
MagicMimeIdentifier.register(RarArchive, "application/vnd.rar")
MagicDescriptionIdentifier.register(RarArchive, lambda s: "rar archive" in s.lower())



import asyncio
import logging
import os
import tempfile
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Type, Union, cast

from ofrak_patch_maker.toolchain.abstract import Toolchain
from ofrak.component.modifier import Modifier
from ofrak.core.architecture import ProgramAttributes
from ofrak.core.complex_block import ComplexBlock
from ofrak.core.injector import BinaryInjectorModifier, BinaryInjectorModifierConfig
from ofrak.core.memory_region import MemoryRegion
from ofrak.core.program import Program
from ofrak.model.component_model import ComponentConfig
from ofrak.resource import Resource
from ofrak.service.resource_service_i import ResourceFilter, ResourceSort, ResourceSortDirection
from ofrak.core.patch_maker.linkable_binary import LinkableBinary
from ofrak_patch_maker.model import PatchRegionConfig, FEM
from ofrak_patch_maker.patch_maker import PatchMaker
from ofrak_patch_maker.toolchain.abstract import Toolchain
from ofrak_patch_maker.toolchain.model import Segment, ToolchainConfig
from ofrak_type.memory_permissions import MemoryPermissions

LOGGER = logging.getLogger(__file__)


class SourceBundle(Dict[str, Union[bytes, "SourceBundle"]]):
    """
    Class used to store filesystem trees of source code as serializable in-memory trees, for
    transfer between components.
    """

    @classmethod
    def slurp(cls, path: str) -> "SourceBundle":
        """
        Slurp up a path into a SourceBundle, recursively getting all files and directories and
        storing them as a tree in memory.

        :param path:
        :return:
        """
        root, dirs, files = next(os.walk(path, topdown=True))

        pairs: List[Tuple[str, Union[bytes, SourceBundle]]] = []
        for file_name in files:
            file_path = os.path.join(root, file_name)
            with open(file_path, "rb") as f:
                file_contents = f.read()

            pairs.append((file_name, file_contents))

        for dir_name in dirs:
            dir_path = os.path.join(root, dir_name)
            pairs.append((dir_name, SourceBundle.slurp(dir_path)))

        return cls(pairs)

    def dump(self, target_path: str):
        """
        Dump a SourceBundle tree back into the local filesystem, at the given target path.
        :param target_path:
        :return:
        """
        os.makedirs(target_path, exist_ok=True)
        for item_name, item_contents in self.items():
            item_path = os.path.join(target_path, item_name)
            if type(item_contents) is bytes:
                # item is a file
                with open(item_path, "wb") as f:
                    f.write(item_contents)
            else:
                # item is a directory
                cast(SourceBundle, item_contents).dump(item_path)


@dataclass
class PatchFromSourceModifierConfig(ComponentConfig):
    """
    :var source_code: Path to directory containing source code (ideally ONLY source code)
    :var source_patches: path of each source file to build and inject, with one or more segments
    defining where to inject one or more of the .text, .data, and .rodata from the build file
    :var toolchain_config: configuration for the
    [Toolchain][ofrak_patch_maker.toolchain.abstract.Toolchain] to use
    :var toolchain: the type of which [Toolchain][ofrak_patch_maker.toolchain.abstract.Toolchain]
      to use to build patch
    :var header_directories: (Optional) paths to directories to search for header files in
    :var patch_name: Optional name of patch
    """

    source_code: SourceBundle
    source_patches: Dict[str, Tuple[Segment, ...]]
    toolchain_config: ToolchainConfig
    toolchain: Type[Toolchain]
    header_directories: Tuple[SourceBundle, ...] = ()
    patch_name: Optional[str] = None


class PatchFromSourceModifier(Modifier):
    """
    Modifier exposing some basic source code patching capabilities.
    """

    targets = (Program,)

    async def modify(self, resource: Resource, config: PatchFromSourceModifierConfig) -> None:
        if config.patch_name is None:
            patch_name = f"{resource.get_id().hex()}_patch"
        else:
            patch_name = config.patch_name

        build_tmp_dir = tempfile.mkdtemp()

        source_tmp_dir = tempfile.mkdtemp()
        config.source_code.dump(source_tmp_dir)

        header_dirs = []
        for header_directory in config.header_directories:
            header_tmp_dir = tempfile.mkdtemp()
            header_directory.dump(header_tmp_dir)
            header_dirs.append(header_tmp_dir)

        absolute_source_list = [
            os.path.join(source_tmp_dir, src_file) for src_file in config.source_patches.keys()
        ]
        program_attributes = await resource.analyze(ProgramAttributes)
        patch_maker = PatchMaker(
            toolchain=config.toolchain(program_attributes, config.toolchain_config),
            build_dir=build_tmp_dir,
        )

        patch_bom = patch_maker.make_bom(
            name=patch_name,
            source_list=absolute_source_list,
            object_list=[],
            header_dirs=header_dirs,
        )

        # Map each object file in the BOM to the segments associated with its source file
        patch_bom_segment_mapping = {
            patch_bom.object_map[os.path.join(source_tmp_dir, src_file)].path: src_segments
            for src_file, src_segments in config.source_patches.items()
        }

        target_program = await resource.view_as(Program)
        target_linkable_bom_info = await target_program.make_linkable_bom(
            patch_maker,
            build_tmp_dir,
            patch_bom.unresolved_symbols,
        )

        # To support additional dynamic references in user space executables
        # Create and use a modifier that will:
        # 1. Extend .got, add new entry
        # 2. Extend .got.plt, add new stub code
        # 3. If the DSO is not already listed in the load list for executable it must be extended and added.
        # 4. Provide the additional .got and .got.plt symbols to make_fem now that we have the locations
        # NOTE: These external functions will probably be *UND*
        p = PatchRegionConfig(patch_bom.name + "_patch", patch_bom_segment_mapping)
        exec_path = os.path.join(build_tmp_dir, "output_exec")
        fem = patch_maker.make_fem(
            [(patch_bom, p), target_linkable_bom_info],
            exec_path,
        )

        await resource.run(
            SegmentInjectorModifier,
            SegmentInjectorModifierConfig.from_fem(fem),
        )

        # Refresh LinkableBinary with the LinkableSymbols used in this patch
        target_binary = await resource.view_as(LinkableBinary)
        await target_binary.define_linkable_symbols_from_patch(
            fem.executable.symbols, program_attributes
        )


@dataclass
class SegmentInjectorModifierConfig(ComponentConfig):
    segments_and_data: Tuple[Tuple[Segment, bytes], ...]

    @staticmethod
    def from_fem(fem: FEM) -> "SegmentInjectorModifierConfig":
        """
        Automatically build a config from a FEM by extracting each segment's bytes and metadata.
        """
        extracted_segments: List[Tuple[Segment, bytes]] = []
        with open(fem.executable.path, "rb") as f:
            exe_data = f.read()
        for segment in fem.executable.segments:
            if segment.length == 0:
                continue
            segment_data = exe_data[segment.offset : segment.offset + segment.length]
            extracted_segments.append((segment, segment_data))
        return SegmentInjectorModifierConfig(tuple(extracted_segments))


class SegmentInjectorModifier(Modifier[SegmentInjectorModifierConfig]):
    """
    Inject some segments into a Program resource. Only segments with non-zero length are
    injected, excluding .bss.
    """

    targets = (Program,)

    async def modify(self, resource: Resource, config: SegmentInjectorModifierConfig) -> None:
        sorted_regions = list(
            await resource.get_descendants_as_view(
                MemoryRegion,
                r_filter=ResourceFilter(
                    include_self=True,
                    tags=(MemoryRegion,),
                ),
                r_sort=ResourceSort(
                    attribute=MemoryRegion.Size,
                    direction=ResourceSortDirection.DESCENDANT,
                ),
            )
        )

        injection_tasks: List[Tuple[Resource, BinaryInjectorModifierConfig]] = []

        for segment, segment_data in config.segments_and_data:
            if segment.length == 0 or segment.vm_address == 0:
                continue
            if segment.length > 0:
                LOGGER.debug(
                    f"    Segment {segment.segment_name} - {segment.length} "
                    f"bytes @ {hex(segment.vm_address)}",
                )
            if segment.segment_name.startswith(".bss"):
                continue
            if segment.segment_name.startswith(".rela"):
                continue
            if segment.segment_name.startswith(".got"):
                # Create new .got and .plt in the exec format here here once we begin supporting
                # the addition of new dynamic references in our patches.
                # For instance, a new call to kmalloc that never existed before.
                # See PatchFromSourceModifier
                continue

            patches = [(segment.vm_address, segment_data)]
            region = MemoryRegion.get_mem_region_with_vaddr_from_sorted(
                segment.vm_address, sorted_regions
            )
            if region is None:
                raise ValueError(
                    f"Cannot inject patch because the memory region at vaddr "
                    f"{hex(segment.vm_address)} is None"
                )

            injection_tasks.append((region.resource, BinaryInjectorModifierConfig(patches)))

        for injected_resource, injection_config in injection_tasks:
            result = await injected_resource.run(BinaryInjectorModifier, injection_config)
            # The above can patch data of any of injected_resources' descendants or ancestors
            # We don't want to delete injected_resources or its ancestors, so subtract them from the
            # set of patched resources
            patched_descendants = result.resources_modified.difference(
                {
                    r.get_id()
                    for r in await injected_resource.get_ancestors(
                        ResourceFilter(include_self=True)
                    )
                }
            )
            to_delete = [
                r for r in await resource.get_descendants() if r.get_id() in patched_descendants
            ]
            await asyncio.gather(*(r.delete() for r in to_delete))


@dataclass
class FunctionReplacementModifierConfig(ComponentConfig):
    """
    :var source_code: Path to directory containing source code (ideally ONLY source code)
    :var new_function_sources: a mapping from function names (to replace) to source code file paths (to use as
    replacements). The paths are relative paths within the source code FilesystemRoot.
    :var toolchain_config: configuration for the
    [Toolchain][ofrak_patch_maker.toolchain.abstract.Toolchain] to use
    :var toolchain: the type of which type of [Toolchain][ofrak_patch_maker.toolchain.abstract.Toolchain] to use
    :var patch_name: Optional name of patch
    :var header_directories: (Optional) paths to directories to search for header files in
    """

    source_code: SourceBundle
    new_function_sources: Dict[str, str]
    toolchain_config: ToolchainConfig
    toolchain: Type[Toolchain]
    patch_name: Optional[str] = None
    header_directories: Tuple[SourceBundle, ...] = ()


class FunctionReplacementModifier(Modifier[FunctionReplacementModifierConfig]):
    """
    Replace one or several functions in a `Program` resource.

    It takes a mapping from function names (to replace) to source code file paths (to use as replacements), then
    for each function to replace, creates a `Segment` from the corresponding complex block and uses the
    `PatchFromSourceModifier` to overwrite this segment with the code taken from the replacement source code file.
    """

    targets = (Program,)

    async def modify(self, resource: Resource, config: FunctionReplacementModifierConfig) -> None:
        program = await resource.view_as(Program)
        function_to_replace_cbs = {
            func_name: await program.get_function_complex_block(func_name)
            for func_name in config.new_function_sources.keys()
        }
        await self._verify_modes_are_the_same(list(function_to_replace_cbs.values()))
        source_patches: Dict[str, Tuple[Segment, ...]] = {
            config.new_function_sources[func_name]: (self._make_text_segment(complex_block),)
            for func_name, complex_block in function_to_replace_cbs.items()
        }
        patch_from_source_config = PatchFromSourceModifierConfig(
            config.source_code,
            source_patches,
            config.toolchain_config,
            config.toolchain,
            config.header_directories,
            config.patch_name,
        )
        await resource.run(PatchFromSourceModifier, patch_from_source_config)

    @staticmethod
    def _make_text_segment(complex_block: ComplexBlock) -> Segment:
        """Return a new code `Segment` corresponding to `complex_block`."""
        return Segment(
            segment_name=".text",
            vm_address=complex_block.virtual_address,
            offset=0,
            is_entry=False,
            length=complex_block.size,
            access_perms=MemoryPermissions.RX,
        )

    @staticmethod
    async def _verify_modes_are_the_same(complex_blocks: List[ComplexBlock]) -> None:
        """
        Verify that the `InstructionSetMode` of all the `complex_blocks` is the same.

        :raises NotImplementedError: if several `InstructionSetMode` values are found
        """
        modes = {await complex_block.get_mode() for complex_block in complex_blocks}
        if len(modes) > 1:
            raise NotImplementedError(
                f"Several values of InstructionSetMode found in complex blocks {complex_blocks}: {modes}\n"
                "This is not currently supported by this component: all complex blocks must have the same mode "
                "in order to be processed in the same patch."
            )
